# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy
import claripy.backends


class TestExpression(unittest.TestCase):
    def test_smudging(self):
        x = claripy.BVS("x", 32)
        y = x + 1
        self.assertTrue(isinstance(y.args[1], claripy.ast.BV))
        self.assertEqual(y.args[1].args[0], 1)
        self.assertEqual(y.args[1].args[1], 32)

        x = claripy.BVS("x", 32)
        y = x * 1
        z = y + 1
        self.assertTrue(isinstance(y.args[1], claripy.ast.BV))
        self.assertEqual(y.args[1].args[0], 1)
        self.assertEqual(y.args[1].args[1], 32)

        self.assertTrue(isinstance(z.args[1], claripy.ast.BV))
        self.assertEqual(z.args[1].args[0], 1)
        self.assertEqual(z.args[1].args[1], 32)

        ccc = claripy.If(x > 10, x * 3 + 2, x * 4 + 2)
        self.assertTrue(isinstance(ccc.args[1].args[1], claripy.ast.BV))
        self.assertEqual(ccc.args[1].args[1].args[0], 2)
        self.assertEqual(ccc.args[1].args[1].args[1], 32)

        x = claripy.BVS("x", 32)
        y = x + "AAAA"
        self.assertTrue(isinstance(y.args[1], claripy.ast.BV))
        self.assertEqual(y.args[1].args[0], 0x41414141)
        self.assertEqual(y.args[1].args[1], 32)

    def test_expression(self):
        bc = claripy.backends.concrete

        e = claripy.BVV(0x01020304, 32)
        self.assertEqual(len(e), 32)
        r = e.reversed
        self.assertEqual(bc.convert(r), 0x04030201)
        self.assertEqual(len(r), 32)

        self.assertEqual([bc.convert(i) for i in r.chop(8)], [4, 3, 2, 1])

        e1 = r[31:24]
        self.assertEqual(bc.convert(e1), 0x04)
        self.assertEqual(len(e1), 8)
        self.assertEqual(bc.convert(e1[2]), 1)
        self.assertEqual(bc.convert(e1[1]), 0)

        ee1 = e1.zero_extend(8)
        self.assertEqual(bc.convert(ee1), 0x0004)
        self.assertEqual(len(ee1), 16)

        ee1 = claripy.BVV(0xFE, 8).sign_extend(8)
        self.assertEqual(bc.convert(ee1), 0xFFFE)
        self.assertEqual(len(ee1), 16)

        xe1 = [bc.convert(i) for i in e1.chop(1)]
        self.assertEqual(xe1, [0, 0, 0, 0, 0, 1, 0, 0])

        a = claripy.BVV(1, 1)
        self.assertEqual(bc.convert(a + a), 2)

        x = claripy.BVV(1, 32)
        self.assertEqual(x.length, 32)
        y = claripy.LShR(x, 10)
        self.assertEqual(y.length, 32)

        r = claripy.BVV(0x01020304, 32)
        rr = r.reversed
        rrr = rr.reversed
        # self.assertIs(bc.convert(r), bc.convert(rrr))
        # self.assertIs(type(bc.convert(rr)), claripy.A)
        self.assertEqual(bc.convert(rr), 0x04030201)
        self.assertIs(r.concat(rr), claripy.Concat(r, rr))

        rsum = r + rr
        self.assertEqual(bc.convert(rsum), 0x05050505)

        r = claripy.BVS("x", 32)
        rr = r.reversed
        rrr = rr.reversed
        self.assertIs(r, rrr)

        # test identity
        self.assertIs(r, rrr)
        self.assertIsNot(r, rr)
        ii = claripy.BVS("ii", 32)
        ij = claripy.BVS("ij", 32)
        self.assertIs(ii, ii)
        self.assertIsNot(ii, ij)

        si = claripy.SI(bits=32, stride=2, lower_bound=20, upper_bound=100)
        sj = claripy.SI(bits=32, stride=2, lower_bound=10, upper_bound=10)
        sk = claripy.SI(bits=32, stride=2, lower_bound=20, upper_bound=100)
        self.assertTrue(claripy.backends.vsa.identical(si, si))
        self.assertFalse(claripy.backends.vsa.identical(si, sj))
        self.assertTrue(claripy.backends.vsa.identical(si, sk))
        self.assertIsNot(si, sj)
        self.assertIsNot(sj, sk)
        self.assertIsNot(sk, si)

        # test hash cache
        self.assertIs(a + a, a + a)

        # test replacement
        old = claripy.BVS("old", 32, explicit_name=True)
        new = claripy.BVS("new", 32, explicit_name=True)
        ooo = claripy.BVV(0, 32)

        old_formula = claripy.If((old + 1) % 256 == 0, old + 10, old + 20)
        new_formula = claripy.replace(old_formula, old, new)
        ooo_formula = claripy.replace(new_formula, new, ooo)

        self.assertNotEqual(old_formula.hash(), new_formula.hash())
        self.assertNotEqual(old_formula.hash(), ooo_formula.hash())
        self.assertNotEqual(new_formula.hash(), ooo_formula.hash())

        self.assertEqual(old_formula.variables, frozenset(("old",)))
        self.assertEqual(new_formula.variables, frozenset(("new",)))
        self.assertEqual(ooo_formula.variables, ooo.variables)

        self.assertTrue(old_formula.symbolic)
        self.assertTrue(new_formula.symbolic)
        self.assertTrue(new_formula.symbolic)

        self.assertEqual(str(old_formula).replace("old", "new"), str(new_formula))
        self.assertEqual(bc.convert(ooo_formula), 20)

        # test dict replacement
        old = claripy.BVS("old", 32, explicit_name=True)
        new = claripy.BVS("new", 32, explicit_name=True)
        c = (old + 10) - (old + 20)
        d = (old + 1) - (old + 2)
        cr = claripy.replace_dict(c, {(old + 10).hash(): (old + 1), (old + 20).hash(): (old + 2)})
        self.assertIs(cr, d)

        # test AST collapse
        s = claripy.SI(bits=32, stride=0, lower_bound=10, upper_bound=10)
        b = claripy.BVV(20, 32)

        sb = s + b
        self.assertIsInstance(sb.args[0], claripy.ast.Base)

        bb = b + b
        # this was broken previously -- it was checking if type(bb.args[0]) == A,
        # and it wasn't, but was instead a subclass. leaving this out for now
        # self.assertNotIsInstance(bb.args[0], claripy.ast.Base)

        # ss = s+s
        # (see above)
        # self.assertNotIsInstance(ss.args[0], claripy.ast.Base)

        sob = s | b
        # for now, this is collapsed. Presumably, Fish will make it not collapse at some point
        self.assertIsInstance(sob.args[0], claripy.ast.Base)

        # make sure the AST collapses for delayed ops like reversing
        rb = b.reversed
        # self.assertIsInstance(rb.args[0], claripy.ast.Base)
        # TODO: Properly delay reversing: should not be eager

        self.assertIsNot(rb, bb)
        self.assertIs(rb, rb)

        # test some alternate bvv creation methods
        self.assertIs(claripy.BVV("AAAA"), claripy.BVV(0x41414141, 32))
        self.assertIs(claripy.BVV("AAAA", 32), claripy.BVV(0x41414141, 32))
        self.assertIs(claripy.BVV("AB"), claripy.BVV(0x4142, 16))
        self.assertIs(claripy.BVV("AB", 16), claripy.BVV(0x4142, 16))
        with self.assertRaises(claripy.errors.ClaripyValueError):
            claripy.BVV("AB", 8)

    def test_cardinality(self):
        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32).annotate(claripy.annotation.StridedIntervalAnnotation(1, 100, 120))
        n = claripy.BVV(10, 32)
        m = claripy.BVV(20, 32)

        self.assertEqual(y.cardinality, 21)
        self.assertEqual(x.cardinality, 2**32)
        self.assertEqual(n.cardinality, 1)
        self.assertEqual(m.cardinality, 1)
        self.assertEqual(n.union(m).cardinality, 2)
        self.assertEqual(n.union(y).cardinality, 111)
        self.assertEqual(y.intersection(x).cardinality, 21)
        self.assertEqual(n.intersection(m).cardinality, 0)
        self.assertEqual(y.intersection(m).cardinality, 0)

        self.assertTrue(n.singlevalued)
        self.assertFalse(n.multivalued)

        self.assertTrue(y.multivalued)
        self.assertFalse(y.singlevalued)

        self.assertFalse(x.singlevalued)
        self.assertTrue(x.multivalued)

        self.assertFalse(y.union(m).singlevalued)
        self.assertTrue(y.union(m).multivalued)

        self.assertFalse(y.intersection(m).singlevalued)
        self.assertFalse(y.intersection(m).multivalued)

    def test_if_stuff(self):
        x = claripy.BVS("x", 32)

        c = claripy.If(x > 10, (claripy.If(x > 10, x * 3, x * 2)), x * 4) + 2
        cc = claripy.If(x > 10, x * 3, x * 4) + 2
        ccc = claripy.If(x > 10, x * 3 + 2, x * 4 + 2)
        cccc = x * claripy.If(x > 10, claripy.BVV(3, 32), claripy.BVV(4, 32)) + 2

        self.assertIs(c, cc)
        self.assertIs(claripy.excavate_ite(c), ccc)
        self.assertIs(claripy.burrow_ite(ccc), cccc)

        i = c + c
        ii = claripy.If(x > 10, (x * 3 + 2) + (x * 3 + 2), (x * 4 + 2) + (x * 4 + 2))
        self.assertIs(claripy.excavate_ite(i), ii)

        cn = claripy.If(x <= 10, claripy.BVV(0x10, 32), 0x20)
        iii = c + cn
        iiii = claripy.If(x > 10, (x * 3 + 2) + 0x20, (x * 4 + 2) + 0x10)
        self.assertIs(claripy.excavate_ite(iii), iiii)

    def test_ite_Solver(self):
        self.raw_ite(claripy.Solver)

    def test_ite_SolverHybrid(self):
        self.raw_ite(claripy.SolverHybrid)

    def test_ite_SolverComposite(self):
        self.raw_ite(claripy.SolverComposite)

    def raw_ite(self, solver_type):
        s = solver_type()
        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        z = claripy.BVS("z", 32)

        ite = claripy.ite_dict(x, {1: 11, 2: 22, 3: 33, 4: 44, 5: 55, 6: 66, 7: 77, 8: 88, 9: 99}, claripy.BVV(0, 32))
        self.assertEqual(sorted(s.eval(ite, 100)), [0, 11, 22, 33, 44, 55, 66, 77, 88, 99])
        # calling it a second time should not fail. it was failing on master due to we incorrectly putting the AST
        # into the eval-exhausted set.
        self.assertEqual(sorted(s.eval(ite, 100)), [0, 11, 22, 33, 44, 55, 66, 77, 88, 99])

        ss = s.branch()
        ss.add(ite == 88)
        self.assertEqual(sorted(ss.eval(ite, 100)), [88])
        self.assertEqual(sorted(ss.eval(x, 100)), [8])

        ity = claripy.ite_dict(x, {1: 11, 2: 22, 3: y, 4: 44, 5: 55, 6: 66, 7: 77, 8: 88, 9: 99}, claripy.BVV(0, 32))
        ss = s.branch()
        ss.add(ity != 11)
        ss.add(ity != 22)
        ss.add(ity != 33)
        ss.add(ity != 44)
        ss.add(ity != 55)
        ss.add(ity != 66)
        ss.add(ity != 77)
        ss.add(ity != 88)
        ss.add(ity != 0)
        ss.add(y == 123)
        self.assertEqual(sorted(ss.eval(ity, 100)), [99, 123])
        self.assertEqual(sorted(ss.eval(x, 100)), [3, 9])
        self.assertEqual(sorted(ss.eval(y, 100)), [123])

        itz = claripy.ite_cases(
            [
                (claripy.And(x == 10, y == 20), 33),
                (claripy.And(x == 1, y == 2), 3),
                (claripy.And(x == 100, y == 200), 333),
            ],
            claripy.BVV(0, 32),
        )
        ss = s.branch()
        ss.add(z == itz)
        ss.add(itz != 0)
        self.assertEqual(ss.eval(y // x, 100), (2,))
        self.assertEqual(sorted(ss.eval(x, 100)), [1, 10, 100])
        self.assertEqual(sorted(ss.eval(y, 100)), [2, 20, 200])

    def test_ite_reverse(self):
        a = claripy.BVS("a", 32)
        cases = [(a == i, claripy.BVV(i, 32)) for i in range(30)]
        ast = claripy.ite_cases(cases, -1)
        ext_cases = list(claripy.reverse_ite_cases(ast))

        assert sum(1 if case.op == "And" else 0 for case, _ in ext_cases)
        for case, val in ext_cases:
            if case.op == "And":
                assert claripy.is_true(val == -1)
            else:
                assert any(case is orig_case and val is orig_val for orig_case, orig_val in cases)

    def test_bool(self):
        bc = claripy.backends.concrete

        a = claripy.And(*[False, False, True])
        self.assertEqual(bc.convert(a), False)
        a = claripy.And(*[True, True, True])
        self.assertEqual(bc.convert(a), True)

        o = claripy.Or(*[False, False, True])
        self.assertEqual(bc.convert(o), True)
        o = claripy.Or(*[False, False, False])
        self.assertEqual(bc.convert(o), False)

    def test_extract(self):
        a = claripy.BVS("a", 32)
        assert a[7:] is a[7:0]
        assert a[31:] is a
        assert a[:] is a
        assert a[:0] is a
        assert a[:-8] is a[31:24]
        assert a[-1:] is a[31:0]
        assert a[-1:-8] is a[31:24]

    def test_get_byte(self):
        a = claripy.BVV("ABCD")
        assert a.get_byte(1) is claripy.BVV("B")

    def test_extract_concat_simplify(self):
        a = claripy.BVS("a", 32)
        assert a[31:0] is a
        assert a[31:8].concat(a[7:0]) is a  # pylint:disable=no-member
        assert a[31:16].concat(a[15:8], a[7:0]) is a  # pylint:disable=no-member
        assert a[31:24].concat(a[23:16], a[15:8], a[7:0]) is a  # pylint:disable=no-member

        a = claripy.BVS("a", 32)
        b = a + 100
        b_concat = b[31:8].concat(b[7:0])
        a100 = a + 100
        assert claripy.is_false(b_concat == a100) is False
        assert list(claripy.Solver().eval(b_concat == a100, 2)) == [True]
        assert b_concat is a100
        assert claripy.is_true(b_concat == a100)

    def test_true_false_cache(self):
        a = claripy.BVS("a_WILL_BE_VIOLATED", 32)
        c = a == a + 1
        assert claripy.backends.z3.is_false(c)
        c.args[1].args = (a, claripy.BVV(0, 32))
        assert claripy.backends.z3.is_false(c)
        assert not claripy.backends.z3.is_true(c)
        assert not claripy.backends.z3.is_false(a == a)

    def test_depth_repr(self):
        x = claripy.BVS("x", 32)
        y = claripy.LShR(x, 10)
        y = claripy.LShR(y, 10)
        y = claripy.LShR(y, 10)
        y = claripy.LShR(y, 10)
        y = claripy.LShR(y, 10)
        y = claripy.LShR(y, 10)
        y = claripy.LShR(y, 10)
        self.assertEqual(
            y.shallow_repr(max_depth=5),
            "<BV32 LShR(LShR(LShR(LShR(LShR(<...>, <...>), 0xa), 0xa), 0xa), 0xa)>",
        )

    def test_rename(self):
        x1 = claripy.BVS("x", 32)
        x2 = x1._rename("y")
        assert x2.variables == frozenset(("y",))

    def test_canonical(self):
        x1 = claripy.BVS("x", 32)
        b1 = claripy.BoolS("b")
        c1 = claripy.BoolS("c")
        x2 = claripy.BVS("x", 32)
        b2 = claripy.BoolS("b")
        c2 = claripy.BoolS("c")

        assert x1.canonicalize()[-1] is x2.canonicalize()[-1]

        y1 = claripy.If(claripy.And(b1, c1), x1, ((x1 + x1) * x1) + 1)
        y2 = claripy.If(claripy.And(b2, c2), x2, ((x2 + x2) * x2) + 1)

        one_names = frozenset.union(x1.variables, b1.variables, c1.variables)
        two_names = frozenset.union(x2.variables, b2.variables, c2.variables)

        assert frozenset.union(*[a.variables for a in y1.leaf_asts()]) == one_names
        assert frozenset.union(*[a.variables for a in y2.leaf_asts()]) == two_names
        assert y1.canonicalize()[-1] is y2.canonicalize()[-1]

    def test_depth(self):
        x1 = claripy.BVS("x", 32)
        assert x1.depth == 1
        x2 = x1 + 1
        assert x2.depth == 2

    def test_multiarg(self):
        x = claripy.BVS("x", 32)
        o = claripy.BVV(2, 32)

        x_add = x + x + x + x
        x_mul = x * x * x * x
        x_sub = x - (x + 1) - (x + 2) - (x + 3)
        x_or = x | (x + 1) | (x + 2) | (x + 3)
        x_xor = x ^ (x + 1) ^ (x + 2) ^ (x + 3)
        x_and = x & (x + 1) & (x + 2) & (x + 3)

        assert x_add.variables == x.variables
        assert x_mul.variables == x.variables
        assert x_sub.variables == x.variables
        assert x_or.variables == x.variables
        assert x_xor.variables == x.variables
        assert x_and.variables == x.variables
        assert (claripy.BVV(1, 32) + (x + x)).variables == x.variables

        assert len(x_add.args) == 4
        assert len(x_mul.args) == 4
        # assert len(x_sub.args) == 4 # needs more work
        assert len(x_or.args) == 4
        assert len(x_xor.args) == 4
        assert len(x_and.args) == 4

        assert claripy.replace(x_add, x, o).args[0] == 8
        assert claripy.replace(x_mul, x, o).args[0] == 16
        assert claripy.replace(x_or, x, o).args[0] == 7
        assert claripy.replace(x_xor, x, o).args[0] == 0
        assert claripy.replace(x_and, x, o).args[0] == 0
        assert (100 + claripy.replace(x_sub, x, o)).args[0] == 90

        # make sure that z3 and vsa backends handle this properly
        claripy.backends.z3.convert(x + x + x + x)
        claripy.backends.vsa.convert(x + x + x + x)

    def test_signed_concrete(self):
        bc = claripy.backends.concrete
        a = claripy.BVV(5, 32)
        b = claripy.BVV(-5, 32)
        c = claripy.BVV(3, 32)
        d = claripy.BVV(-3, 32)

        # test unsigned
        assert bc.convert(a // c) == 1
        assert bc.convert(a // d) == 0
        assert bc.convert(b // c) == 0x55555553
        assert bc.convert(b // d) == 0
        assert bc.convert(a % c) == 2
        assert bc.convert(a % d) == 5
        assert bc.convert(b % c) == 2
        assert bc.convert(b % d) == -5

        # test unsigned
        assert bc.convert(a.SDiv(c)) == 1
        assert bc.convert(a.SDiv(d)) == -1
        assert bc.convert(b.SDiv(c)) == -1
        assert bc.convert(b.SDiv(d)) == 1
        assert bc.convert(a.SMod(c)) == 2
        assert bc.convert(a.SMod(d)) == 2
        assert bc.convert(b.SMod(c)) == -2
        assert bc.convert(b.SMod(d)) == -2

    def test_signed_symbolic(self):
        solver = claripy.Solver()
        a = claripy.BVS("a", 32)
        b = claripy.BVS("b", 32)
        c = claripy.BVS("c", 32)
        d = claripy.BVS("d", 32)
        solver.add(a == 5)
        solver.add(b == -5)
        solver.add(c == 3)
        solver.add(d == -3)

        # test unsigned
        assert list(solver.eval(a // c, 2)) == [1]
        assert list(solver.eval(a // d, 2)) == [0]
        assert list(solver.eval(b // c, 2)) == [0x55555553]
        assert list(solver.eval(b // d, 2)) == [0]
        assert list(solver.eval(a % c, 2)) == [2]
        assert list(solver.eval(a % d, 2)) == [5]
        assert list(solver.eval(b % c, 2)) == [2]
        assert list(solver.eval(b % d, 2)) == [2**32 - 5]

        # test unsigned
        assert list(solver.eval(a.SDiv(c), 2)) == [1]
        assert list(solver.eval(a.SDiv(d), 2)) == [2**32 - 1]
        assert list(solver.eval(b.SDiv(c), 2)) == [2**32 - 1]
        assert list(solver.eval(b.SDiv(d), 2)) == [1]
        assert list(solver.eval(a.SMod(c), 2)) == [2]
        assert list(solver.eval(a.SMod(d), 2)) == [2]
        assert list(solver.eval(b.SMod(c), 2)) == [2**32 - 2]
        assert list(solver.eval(b.SMod(d), 2)) == [2**32 - 2]

    def test_arith_shift(self):
        bc = claripy.backends.concrete
        a = claripy.BVV(-4, 32)
        assert bc.convert(a << 1) == -8
        assert bc.convert(a >> 1) == -2
        assert bc.convert(a << 32) == 0
        assert bc.convert(a >> 32) == -1
        # neg shift is treated as unsigned
        assert bc.convert(a << -1) == bc.convert(a << (2**32 - 1))
        assert bc.convert(a >> -1) == bc.convert(a >> (2**32 - 1))

        solver = claripy.Solver()
        a = claripy.BVS("a", 32)
        solver.add(a == -4)
        assert list(solver.eval(a << 1, 2)) == [2**32 - 8]
        assert list(solver.eval(a >> 1, 2)) == [2**32 - 2]
        assert list(solver.eval(a << 32, 2)) == [0]
        assert list(solver.eval(a >> 32, 2)) == [2**32 - 1]

        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        assert list(solver.eval(a << -x == a << (2**32 - x), 2)) == [True]
        assert list(solver.eval(a >> -x == a >> (2**32 - x), 2)) == [True]
        solver.add(y >= 32)
        assert list(solver.eval(x << y, 2)) == [0]
        assert list(solver.eval(x >> y, 2)) == [0, 2**32 - 1]

    def test_logic_shift_right(self):
        bc = claripy.backends.concrete
        a = claripy.BVV(-4, 32)
        assert bc.convert(a.LShR(1)) == 0x7FFF_FFFE
        assert bc.convert(a.LShR(32)) == 0
        assert bc.convert(a.LShR(-1)) == 0

        solver = claripy.Solver()
        a = claripy.BVS("a", 32)
        solver.add(a == -4)
        assert list(solver.eval(a.LShR(1), 2)) == [0x7FFF_FFFE]

        x = claripy.BVS("x", 32)
        y = claripy.BVS("y", 32)
        solver.add(y >= 32)
        assert list(solver.eval(x.LShR(y), 2)) == [0]

    def test_bool_conversion(self):
        a = claripy.BVV(42, 32)
        try:
            assert a == 42
            assert False, "`assert ast` should raise an exception"
        except claripy.ClaripyOperationError:
            pass

        try:
            bool(a == 42)
            assert False, "bool(ast) should raise an exception"
        except claripy.ClaripyOperationError:
            pass

        try:
            if a == 42:
                pass
            assert False, "`if ast` should raise an exception"
        except claripy.ClaripyOperationError:
            pass

    def test_bool_replace_in_ite(self):
        b = claripy.BoolS("b")
        expr = claripy.If(b, claripy.BVV(2, 32), claripy.BVV(3, 32))
        new_expr = claripy.replace(expr, b, claripy.BoolV(True))

        # Replace calls make_like which will simplify the expression. As a
        # result, the new expression will be a BVV.
        assert new_expr.op == "BVV" and new_expr.args == (2, 32)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import itertools
import logging
import math
import pickle
import struct
from collections import deque
from contextlib import suppress
from enum import IntEnum
from hashlib import blake2b
from itertools import chain
from typing import TYPE_CHECKING, NoReturn, TypeVar, Union, cast
from weakref import WeakValueDictionary

from typing_extensions import Self

import claripy
from claripy import operations
from claripy.errors import BackendError, ClaripyOperationError
from claripy.fp import FSort

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from claripy.annotation import Annotation
    from claripy.backends import Backend

log = logging.getLogger(__name__)

blake2b_unpacker = struct.Struct("Q")

# pylint:disable=unused-argument,too-many-boolean-expressions,too-many-positional-arguments

ArgType = Union["Base", bool, int, float, str, FSort, tuple["ArgType"], None]

T = TypeVar("T", bound="Base")
A = TypeVar("A", bound="Annotation")


class ReprLevel(IntEnum):
    """
    Representation levels for ASTs.
    """

    LITE_REPR = 0
    MID_REPR = 1
    FULL_REPR = 2


#
# AST variable naming
#

var_counter = itertools.count()
_unique_names = True


def _make_name(name: str, size: int, explicit_name: bool = False, prefix: str = "") -> str:
    if _unique_names and not explicit_name:
        return "%s%s_%d_%d" % (prefix, name, next(var_counter), size)
    return name


def _d(h, cls, state):
    """
    This function is the deserializer for ASTs.
    It exists to work around the fact that pickle will (normally) call __new__() with no arguments during
    deserialization. For ASTs, this does not work.
    """
    op, args, length, variables, symbolic, annotations = state
    return cls.__new__(
        cls, op, args, length=length, variables=variables, symbolic=symbolic, annotations=annotations, hash=h
    )


class Base:
    """
    This is the base class of all claripy ASTs. An AST tracks a tree of operations on arguments.

    This class should not be instanciated directly - instead, use one of the constructor functions (BVS, BVV, FPS,
    FPV...) to construct a leaf node and then build more complicated expressions using operations.

    AST objects have *hash identity*. This means that an AST that has the same hash as another AST will be the *same*
    object. This is critical for efficient memory usage. As an example, the following is true::

        a, b = two different ASTs
        c = b + a
        d = b + a
        assert c is d

    :ivar op:                       The operation that is being done on the arguments
    :ivar args:                     The arguments that are being used
    """

    # Hashcons
    op: str
    args: tuple[ArgType, ...]
    length: int | None
    variables: frozenset[str]
    symbolic: bool
    annotations: tuple[Annotation, ...]

    # Derived information
    depth: int
    _hash: int
    _relocatable_annotations: frozenset[Annotation]
    _uneliminatable_annotations: frozenset[Annotation]

    # Backend information
    _errored: set[Backend]

    # Caching
    _cached_encoded_name: bytes | None

    # Extra information
    _uninitialized: bool

    __slots__ = [
        "op",
        "args",
        "length",
        "variables",
        "symbolic",
        "annotations",
        "depth",
        "_hash",
        "_uneliminatable_annotations",
        "_relocatable_annotations",
        "_errored",
        "_cached_encoded_name",
        "_uninitialized",
        "__weakref__",
    ]

    _hash_cache: WeakValueDictionary[int, Base] = WeakValueDictionary()

    def __new__(  # pylint:disable=redefined-builtin
        cls: type[Base],
        op: str,
        args: Iterable[ArgType],
        add_variables: Iterable[str] | None = None,
        hash: int | None = None,
        symbolic: bool | None = None,
        variables: frozenset[str] | None = None,
        errored: set[Backend] | None = None,
        uninitialized: bool = False,
        annotations: tuple[Annotation, ...] = (),
        skip_child_annotations: bool = False,
        length: int | None = None,
        encoded_name: bytes | None = None,
    ):
        """
        This is called when you create a new Base object, whether directly or through an operation.
        It finalizes the arguments (see the _finalize function, above) and then computes
        a hash. If an AST of this hash already exists, it returns that AST. Otherwise,
        it creates, initializes, and returns the AST.

        :param op:                  The AST operation ('__add__', 'Or', etc)
        :param args:                The arguments to the AST operation (i.e., the objects to add)
        :param variables:           The symbolic variables present in the AST (default: empty set)
        :param symbolic:            A flag saying whether or not the AST is symbolic (default: False)
        :param length:              An integer specifying the bit length of this AST (default: None)
        :param simplified:          A measure of how simplified this AST is. 0 means unsimplified,
                                        1 means fast-simplified (basically, just undoing the Reverse
                                        op), and 2 means simplified through z3.
        :param errored:             A set of backends that are known to be unable to handle this AST.
        :param annotations:         A frozenset of annotations applied onto this AST.
        """

        if hash is not None and (self := cls._hash_cache.get(hash, None)) is not None:
            return self

        # if any(isinstance(a, BackendObject) for a in args):
        #   raise Exception('asdf')

        a_args = args if type(args) is tuple else tuple(args)
        b_args = tuple(a for a in a_args if isinstance(a, Base))

        if symbolic is None:
            symbolic = any(a.symbolic for a in b_args)
        if variables is None:
            variables = frozenset(chain.from_iterable(a.variables for a in b_args))
        if errored is None:
            errored = set(chain.from_iterable(a._errored for a in b_args))
        arg_max_depth = max((a.depth for a in b_args), default=0)

        if add_variables:
            variables = frozenset.union(variables, add_variables)

        if not symbolic and op not in operations.leaf_operations:
            with suppress(BackendError):
                r = operations._handle_annotations(
                    claripy.backends.concrete._abstract(claripy.backends.concrete.call(op, args)), args
                )
                if r is not None:
                    return r

        uneliminatable_annotations = frozenset(a for a in annotations if not (a.eliminatable or a.relocatable))
        relocatable_annotations = frozenset(a for a in annotations if not a.eliminatable and a.relocatable)

        if not skip_child_annotations:
            for a in b_args:
                uneliminatable_annotations |= a._uneliminatable_annotations
                relocatable_annotations |= a._relocatable_annotations

            annotations = tuple(frozenset((*annotations, *relocatable_annotations)))

        hash_ = Base._calc_hash(op, a_args, annotations, length)
        self = cls._hash_cache.get(hash_, None)
        if self is None:
            self = super().__new__(cls)
            depth = arg_max_depth + 1
            self.__a_init__(
                op,
                a_args,
                variables,
                symbolic=symbolic,
                length=length,
                errored=errored,
                uninitialized=uninitialized,
                annotations=annotations,
                encoded_name=encoded_name,
                depth=depth,
                uneliminatable_annotations=uneliminatable_annotations,
                relocatable_annotations=relocatable_annotations,
            )
            self._hash = hash_
            cls._hash_cache[hash_] = self
        # else:
        #     if not self._check_args_same(a_args) or self.op != op or self.annotations != annotations:
        #         raise Exception("CRAP -- hash collision")

        return self

    def make_like(
        self,
        op: str,
        args: Iterable[ArgType],
        simplify: bool = False,
        annotations: tuple[Annotation, ...] | None = None,
        variables: frozenset[str] | None = None,
        symbolic: bool | None = None,
        uninitialized: bool = False,
        skip_child_annotations: bool = False,
        length: int | None = None,
    ) -> Self:
        # Try to simplify the expression again
        simplified, annotated = claripy.simplifications.simplify(op, args) if simplify else (None, False)
        if simplified is not None:
            op = simplified.op
        if (  # fast path
            simplified is None
            and annotations
            and variables is None
            and symbolic is None
            and uninitialized is False
            and skip_child_annotations
            and length is not None
        ):
            uneliminatable_annotations = frozenset(
                anno for anno in annotations if not anno.eliminatable and not anno.relocatable
            )
            relocatable_annotations = frozenset(
                anno for anno in annotations if not anno.eliminatable and anno.relocatable
            )

            cache = type(self)._hash_cache
            h = Base._calc_hash(op, args, annotations, length)
            cached_ast = cast(T | None, cache.get(h, None))
            if cached_ast is not None:
                return cached_ast

            result: T = super().__new__(type(self))
            result.__a_init__(
                op,
                args,
                self.variables,
                depth=self.depth,
                uneliminatable_annotations=uneliminatable_annotations,
                relocatable_annotations=relocatable_annotations,
                symbolic=self.symbolic,
                annotations=annotations,
                length=length,
                uninitialized=self._uninitialized,
            )

            result._hash = h
            cache[h] = result

            return result

        all_operations = operations.leaf_operations_symbolic_with_union
        # special case: if self is one of the args, we do not copy annotations over from self since child
        # annotations will be re-processed during AST creation.
        if annotations is None:
            if not annotated:
                annotations = self.annotations if not args or not any(self is arg for arg in args) else ()
            else:
                annotations = simplified.annotations
        if variables is None and op in all_operations:
            variables = self.variables
        if uninitialized is None:
            uninitialized = self._uninitialized
        if symbolic is None and op in all_operations:
            symbolic = self.symbolic

        return type(self)(
            op,
            args if simplified is None else simplified.args,
            annotations=annotations,
            variables=variables,
            uninitialized=uninitialized,
            symbolic=symbolic,
            skip_child_annotations=skip_child_annotations,
            length=length,
        )

    # pylint:disable=attribute-defined-outside-init
    def __a_init__(
        self: Self,
        op: str,
        args: tuple[ArgType, ...],
        variables: frozenset[str],
        symbolic: bool | None = None,
        length: int | None = None,
        errored: set[Backend] | None = None,
        uninitialized: bool = False,
        annotations: tuple[Annotation, ...] | None = None,
        encoded_name: bytes | None = None,
        depth: int | None = None,
        uneliminatable_annotations: frozenset[Annotation] | None = None,
        relocatable_annotations: frozenset[Annotation] | None = None,
    ) -> None:
        """
        Initializes an AST. Takes the same arguments as ``Base.__new__()``

        We use this instead of ``__init__`` due to python's undesirable behavior w.r.t. automatically calling it on
        return from ``__new__``.
        """

        # HASHCONS: these attributes key the cache
        # BEFORE CHANGING THIS, SEE ALL OTHER INSTANCES OF "HASHCONS" IN THIS FILE
        self.op = op
        self.args = args if type(args) is tuple else tuple(args)
        self.length = length
        self.variables = variables
        self.symbolic = symbolic
        self.annotations: tuple[Annotation] = annotations
        self._uneliminatable_annotations = uneliminatable_annotations
        self._relocatable_annotations = relocatable_annotations

        self.depth = depth if depth is not None else 1

        self._cached_encoded_name = encoded_name

        self._errored = errored if errored is not None else set()

        self._uninitialized = uninitialized

        if len(self.args) == 0:
            raise ClaripyOperationError("AST with no arguments!")

    # pylint:enable=attribute-defined-outside-init
    def __reduce__(self):
        # HASHCONS: these attributes key the cache
        # BEFORE CHANGING THIS, SEE ALL OTHER INSTANCES OF "HASHCONS" IN THIS FILE
        return _d, (
            self._hash,
            self.__class__,
            (self.op, self.args, self.length, self.variables, self.symbolic, self.annotations),
        )

    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def _calc_hash(
        op: str,
        args: tuple[Base, ...],
        annotations: tuple[Annotation, ...],
        length: int | None,
    ) -> int:
        """
        Calculates the hash of an AST, given the operation, args, and kwargs.

        :param op:                  The operation.
        :param args:                The args tuple.
        :param annotations:         The annotations tuple.
        :returns:                   a hash, as bytes.

        We do it using md5 to avoid hash collisions.
        (hash(-1) == hash(-2), for example)
        """
        # HASHCONS: these attributes key the cache
        # BEFORE CHANGING THIS, SEE ALL OTHER INSTANCES OF "HASHCONS" IN THIS FILE

        to_hash = Base._ast_serialize(op, args, annotations, length)

        hd = blake2b(to_hash, digest_size=8).digest()
        return blake2b_unpacker.unpack(hd)[0]

    @staticmethod
    def _ast_serialize(
        op: str,
        args: tuple[ArgType, ...],
        annotations: tuple[Annotation, ...],
        length: int | None,
    ) -> bytes:
        """
        Serialize the AST and get a bytestring for hashing.

        :param op:          The operator.
        :param args:        The args tuple.
        :param annotations: The annotations tuple.
        :return:            The serialized bytestring.
        """

        serialized_args = b"".join(b"<" + Base._arg_serialize(a) + b">" for a in args)
        serialized_annotations = b"".join(b"(" + Base._arg_serialize(a) + b")" for a in annotations)
        serialized_length = b"" if length is None else length.to_bytes(8, "little", signed=False)

        return b"{" + op.encode() + serialized_args + serialized_annotations + serialized_length + b"}"

    @staticmethod
    def _arg_serialize(arg: ArgType | Annotation) -> bytes:
        """Serialize one argument."""

        if isinstance(arg, Base):
            return arg._hash.to_bytes(8, "little", signed=False)
        if arg is None:
            return b"\x0f"
        if arg is True:
            return b"\x1f"
        if arg is False:
            return b"\x2e"
        if isinstance(arg, int):
            return arg.to_bytes((arg.bit_length() + 15) // 8, "little", signed=True)
        if isinstance(arg, str):
            return arg.encode()
        if isinstance(arg, float):
            if math.isnan(arg):
                return b"nan"
            if math.isinf(arg):
                return b"inf"
            if arg == 0.0 and math.copysign(1, arg) < 0:
                return b"-0.0"
            return struct.pack("d", arg)
        if isinstance(arg, tuple):
            return b"".join(b"<" + Base._arg_serialize(a) + b">" for a in arg)
        if hasattr(arg, "__hash__"):
            return hash(arg).to_bytes(8, "little", signed=True)

        log.debug("Don't know how to serialize %s, consider implementing __hash__", arg)
        return pickle.dumps(arg)

    def __hash__(self) -> int:
        return self._hash

    def hash(self) -> int:
        """Python's built in hash function is not collision resistant, so we use our own.
        When you call `hash(ast)`, the value is derived from the claripy hash, but it gets
        passed through python's non-resistent hash function first. This skips that step,
        allowing the claripy hash to be used directly, eg as a cache key.
        """
        return self._hash

    @property
    def _encoded_name(self) -> bytes:
        if self._cached_encoded_name is None:
            self._cached_encoded_name = self.args[0].encode()
        return self._cached_encoded_name

    def _check_args_same(self, other_args: tuple[ArgType, ...], lenient_names=False) -> bool:
        """
        Check if two ASTs are the same.
        """
        # Several types inside of args don't support normall == comparison, so if we see those,
        # we need compare them manually.
        for a, b in zip(self.args, other_args, strict=True):
            if isinstance(a, Base) and isinstance(b, Base):
                if a._hash != b._hash:
                    return False
                continue
            if isinstance(a, float) and isinstance(b, float):
                if math.isnan(a) and math.isnan(b):
                    continue
                if math.isinf(a) and math.isinf(b):
                    continue
                if a != b:
                    return False
            if lenient_names and isinstance(a, str) and isinstance(b, str):
                continue
            if a != b:
                return False

        return True

    def identical(self, other: Self, strict=False) -> bool:
        """
        Check if two ASTs are identical. If `strict` is False, the comparison
        will be lenient on the names of the ASTs.
        """
        return self._hash == other._hash or (
            self.op == other.op
            and self._check_args_same(other.args, lenient_names=not strict)
            and self.annotations == other.annotations
        )

    #
    # Collapsing and simplification
    #

    def _rename(self, new_name: str) -> Self:
        if self.op not in {"BVS", "BoolS", "FPS"}:
            raise ClaripyOperationError("rename is only supported on leaf nodes")
        new_args = (new_name,) + self.args[1:]
        return self.make_like(self.op, new_args, length=self.length, variables=frozenset((new_name,)))

    #
    # Annotations
    #

    def _apply_to_annotations(self, f):
        return self.make_like(self.op, self.args, annotations=f(self.annotations), skip_child_annotations=True)

    def has_annotation_type(self, annotation_type: type[Annotation]) -> bool:
        """
        Check if this AST has an annotation of a given type.

        :param annotation_type:     The type of the annotation.
        :return:                    True if the AST has an annotation of the given type.
        """
        return any(isinstance(a, annotation_type) for a in self.annotations)

    def get_annotations_by_type(self, annotation_type: type[A]) -> tuple[A, ...]:
        """
        Get all annotations of a given type.

        :param annotation_type:     The type of the annotation.
        :return:                    A tuple of annotations of the given type.
        """
        return tuple(a for a in self.annotations if isinstance(a, annotation_type))

    def get_annotation(self, annotation_type: type[A]) -> A | None:
        """
        Get the first annotation of a given type.

        :param annotation_type:     The type of the annotation.
        :return:                    The annotation of the given type, or None if not found.
        """
        for a in self.annotations:
            if isinstance(a, annotation_type):
                return a
        return None

    def append_annotation(self, a: Annotation) -> Self:
        """
        Appends an annotation to this AST.

        :param a:                   the annotation to append
        :returns:                   a new AST, with the annotation added
        """
        return self._apply_to_annotations(lambda alist: (*alist, a))

    def append_annotations(self, new_tuple: tuple[Annotation, ...]) -> Self:
        """
        Appends several annotations to this AST.

        :param new_tuple:           the tuple of annotations to append
        :returns:                   a new AST, with the annotations added
        """
        return self._apply_to_annotations(lambda alist: alist + new_tuple)

    def annotate(self, *args: Annotation, remove_annotations: Iterable[Annotation] | None = None) -> Self:
        """
        Appends annotations to this AST.

        :param args:                the tuple of annotations to append (variadic positional args)
        :param remove_annotations:  annotations to remove
        :returns:                   a new AST, with the annotations added
        """
        if not remove_annotations:
            return self._apply_to_annotations(lambda alist: alist + args)
        return self._apply_to_annotations(
            lambda alist: tuple(arg for arg in alist if arg not in remove_annotations) + args
        )

    def insert_annotation(self, a: Annotation) -> Self:
        """
        Inserts an annotation to this AST.

        :param a:                   the annotation to insert
        :returns:                   a new AST, with the annotation added
        """
        return self._apply_to_annotations(lambda alist: (a, *alist))

    def insert_annotations(self, new_tuple: tuple[Annotation, ...]) -> Self:
        """
        Inserts several annotations to this AST.

        :param new_tuple:           the tuple of annotations to insert
        :returns:                   a new AST, with the annotations added
        """
        return self._apply_to_annotations(lambda alist: new_tuple + alist)

    def replace_annotations(self, new_tuple: tuple[Annotation, ...]) -> Self:
        """
        Replaces annotations on this AST.

        :param new_tuple:           the tuple of annotations to replace the old annotations with
        :returns:                   a new AST, with the annotations added
        """
        return self._apply_to_annotations(lambda alist: new_tuple)

    def remove_annotation(self, a: Annotation) -> Self:
        """
        Removes an annotation from this AST.

        :param a:                   the annotation to remove
        :returns:                   a new AST, with the annotation removed
        """
        return self._apply_to_annotations(lambda alist: tuple(oa for oa in alist if oa != a))

    def remove_annotations(self, remove_sequence: Iterable[Annotation]) -> Self:
        """
        Removes several annotations from this AST.

        :param remove_sequence:     a sequence/set of the annotations to remove
        :returns:                   a new AST, with the annotations removed
        """
        return self._apply_to_annotations(lambda alist: tuple(oa for oa in alist if oa not in remove_sequence))

    def clear_annotations(self) -> Self:
        """
        Removes all annotations from this AST.

        :returns:                   a new AST, with all annotations removed
        """
        return self._apply_to_annotations(lambda _: ())

    def clear_annotation_type(self, annotation_type: type[Annotation]) -> Self:
        """
        Removes all annotations of a given type from this AST.

        :param annotation_type:     the type of annotations to remove
        :returns:                   a new AST, with the annotations removed
        """
        return self._apply_to_annotations(
            lambda alist: tuple(oa for oa in alist if not isinstance(oa, annotation_type))
        )

    #
    # Viewing and debugging
    #

    def dbg_repr(self, prefix=None) -> str:  # pylint:disable=unused-argument
        """
        Returns a debug representation of this AST.
        """
        return self.shallow_repr(max_depth=None, details=ReprLevel.FULL_REPR)

    def _type_name(self) -> str:
        return self.__class__.__name__

    def __repr__(self, inner=False, max_depth=None, explicit_length=False):
        return self.shallow_repr(max_depth=max_depth, explicit_length=explicit_length, inner=inner)

    def shallow_repr(
        self,
        max_depth: int = 8,
        explicit_length: bool = False,
        details: ReprLevel = ReprLevel.LITE_REPR,
        inner: bool = False,
        parent_prec: int = 15,
        left: bool = True,
    ) -> str:
        """
        Returns a string representation of this AST, but with a maximum depth to
        prevent floods of text being printed.

        :param max_depth:           The maximum depth to print.
        :param explicit_length:     Print lengths of BVV arguments.
        :param details:             An integer value specifying how detailed the output should be:
                                        LITE_REPR - print short repr for both operations and BVs,
                                        MID_REPR  - print full repr for operations and short for BVs,
                                        FULL_REPR - print full repr of both operations and BVs.
        :param inner:               whether or not it is an inner AST
        :param parent_prec:         parent operation precedence level
        :param left:                whether or not it is a left AST
        :returns:                   A string representing the AST
        """
        if max_depth is not None and max_depth <= 0:
            return "<...>"

        if self.op in operations.reversed_ops:
            op = operations.reversed_ops[self.op]
            args = reversed(self.args)
        else:
            op = self.op
            args = self.args

        next_max_depth = max_depth - 1 if max_depth is not None else None
        length = self.length if explicit_length else None
        # if operation is not in op_precedence, assign the "least operation precedence"
        op_prec = operations.op_precedence.get(op, 15)

        args = [
            (
                arg.shallow_repr(next_max_depth, explicit_length, details, True, op_prec, idx == 0)
                if isinstance(arg, Base)
                else arg
            )
            for idx, arg in enumerate(args)
        ]

        prec_diff = parent_prec - op_prec
        inner_infix_use_par = prec_diff < 0 or prec_diff == 0 and not left
        inner_repr = self._op_repr(op, args, inner, length, details, inner_infix_use_par)

        if not inner:
            return f"<{self._type_name()} {inner_repr}>"
        return inner_repr

    @staticmethod
    def _op_repr(
        op: str,
        args: tuple[ArgType, ...],
        inner: ArgType,
        length: int | None,
        details: ReprLevel,
        inner_infix_use_par: bool,
    ):
        if details < ReprLevel.FULL_REPR:
            if op == "BVS":
                return f"{args[0]}"

            if op == "BoolV":
                return str(args[0])

            if op == "BVV":
                if args[0] is None:
                    value = "!"
                elif args[1] < 10:
                    value = format(args[0], "")
                else:
                    value = format(args[0], "#x")
                return value + "#%d" % length if length is not None else value

        if details < ReprLevel.MID_REPR:
            if op == "If":
                value = f"if {args[0]} then {args[1]} else {args[2]}"
                return f"({value})" if inner else value

            if op == "Not":
                return f"!{args[0]}"

            if op == "Extract":
                return f"{args[2]}[{args[0]}:{args[1]}]"

            if op == "ZeroExt":
                value = f"0#{args[0]} .. {args[1]}"
                return f"({value})" if inner else value

            if op in operations.prefix:
                assert len(args) == 1
                value = f"{operations.prefix[op]}{args[0]}"
                return f"({value})" if inner and inner_infix_use_par else value

            if op in operations.infix:
                value = f" {operations.infix[op]} ".join(args)
                return f"({value})" if inner and inner_infix_use_par else value

        return "{}({})".format(op, ", ".join(str(arg) for arg in args))

    def children_asts(self) -> Iterator[Base]:
        """
        Return an iterator over the nested children ASTs.
        """
        ast_queue = deque([iter(self.args)])
        while ast_queue:
            try:
                ast = next(ast_queue[-1])
            except StopIteration:
                ast_queue.pop()
                continue

            if isinstance(ast, Base):
                ast_queue.append(iter(ast.args))

                log.debug("Yielding AST %s with hash %s with %d children", ast, ast.hash(), len(ast.args))
                yield ast

    def leaf_asts(self) -> Iterator[Base]:
        """
        Return an iterator over the leaf ASTs.
        """
        seen = set()

        ast_queue = deque([self])
        while ast_queue:
            ast = ast_queue.pop()
            if isinstance(ast, Base) and ast.hash() not in seen:
                seen.add(ast.hash())

                if ast.depth == 1:
                    yield ast
                    continue

                ast_queue.extend(ast.args)
                continue

    def is_leaf(self) -> bool:
        """
        Check if this AST is a leaf node.
        """
        return self.depth == 1

    def dbg_is_looped(self) -> Base | bool:  # TODO: this return type is bad
        log.debug("Checking AST with hash %s for looping", self.hash())

        seen = set()
        for child_ast in self.children_asts():
            if child_ast.hash() in seen:
                return child_ast
            seen.add(child_ast.hash())

        return False

    #
    # Various AST modifications (replacements)
    #

    def swap_args(self, new_args: tuple[ArgType, ...], new_length: int | None = None, **kwargs) -> Self:
        """
        This returns the same AST, with the arguments swapped out for new_args.
        """

        if len(self.args) == len(new_args) and all(a is b for a, b in zip(self.args, new_args, strict=False)):
            return self

        length = self.length if new_length is None else new_length
        return self.make_like(self.op, new_args, length=length, **kwargs)

    #
    # Other helper functions
    #

    def split(self, split_on: Iterable[str]) -> list[ArgType]:
        """
        Splits the AST if its operation is `split_on` (i.e., return all the arguments). Otherwise, return a list with
        just the AST.
        """
        if self.op in split_on:
            return list(self.args)
        return [self]

    # we don't support iterating over Base objects
    def __iter__(self) -> NoReturn:
        """
        This prevents people from iterating over ASTs.
        """
        raise ClaripyOperationError("Please don't iterate over, or split, AST nodes!")

    def __bool__(self) -> NoReturn:
        """
        This prevents people from accidentally using an AST as a condition. For
        example, the following was previously common::

            a,b = two ASTs
            if a == b:
                do something

        The problem is that `a == b` would return an AST, because an AST can be symbolic
        and there could be no way to actually know the value of that without a
        constraint solve. This caused tons of issues.
        """
        raise ClaripyOperationError(
            "testing Expressions for truthiness does not do what you want, as these expressions can be symbolic"
        )

    def structurally_match(self, o: Base) -> bool:
        """
        Structurally compares two A objects, and check if their corresponding leaves are definitely the same A object
        (name-wise or hash-identity wise).

        :param o:                   the other claripy A object
        :returns:                   True/False
        """

        # TODO: Convert a and b into canonical forms

        if self.op != o.op:
            return False

        if len(self.args) != len(o.args):
            return False

        for arg_a, arg_b in zip(self.args, o.args, strict=False):
            if not isinstance(arg_a, Base):
                if type(arg_a) != type(arg_b):  # noqa: E721
                    return False
                # They are not ASTs
                if arg_a != arg_b:
                    return False
                continue

            if arg_a.is_leaf():
                if arg_a is not arg_b:
                    return False

            else:
                if not arg_a.structurally_match(arg_b):
                    return False

        return True

    def canonicalize(self, var_map=None, counter=None) -> Self:
        counter = itertools.count() if counter is None else counter
        var_map = {} if var_map is None else var_map

        for v in self.leaf_asts():
            if v.hash() not in var_map and v.op in {"BVS", "BoolS", "FPS"}:
                new_name = "canonical_%d" % next(counter)
                var_map[v.hash()] = v._rename(new_name)

        return var_map, counter, claripy.replace_dict(self, var_map)

    #
    # these are convenience operations
    #

    def _first_backend(self, what):
        for b in claripy.backends.all_backends:
            if b in self._errored:
                continue

            try:
                return getattr(b, what)(self)
            except BackendError:
                pass
        return None

    @property
    def concrete_value(self):
        try:
            raw = claripy.backends.concrete.convert(self)
            if isinstance(raw, bool):
                return raw
            return raw.value
        except BackendError:
            return self

    @property
    def singlevalued(self) -> bool:
        return self._first_backend("singlevalued")

    @property
    def multivalued(self) -> bool:
        return self._first_backend("multivalued")

    @property
    def cardinality(self) -> int:
        return self._first_backend("cardinality")

    @property
    def concrete(self) -> bool:
        return not self.symbolic

    @property
    def uninitialized(self) -> bool:
        """
        Whether this AST comes from an uninitialized dereference or not. It's only used in under-constrained symbolic
        execution mode.

        :returns:                   True/False/None (unspecified).
        """

        # TODO: It should definitely be moved to the proposed Annotation backend.

        return self._uninitialized



from __future__ import annotations

from claripy import false
from claripy.errors import ClaripyValueError, UnsatError


class ConstraintFilterMixin:
    """ConstraintFilterMixin is a mixin that filters True and False constraints."""

    def _constraint_filter(self, constraints):
        if type(constraints) not in (tuple, list):
            raise ClaripyValueError("The extra_constraints argument should be a list of constraints.")

        if len(constraints) == 0:
            return constraints

        ccs = [self._concrete_constraint(c) for c in constraints]
        if False in ccs:
            raise UnsatError("Constraints contain False.")
        return tuple((o if n is None else o) for o, n in zip(constraints, ccs, strict=False) if n is not True)

    def _add(self, constraints, invalidate_cache=True):
        try:
            ec = self._constraint_filter(constraints)
        except UnsatError:
            # filter out concrete False
            ec = [c for c in constraints if c not in {False, false()}] + [false()]

        if len(constraints) == 0:
            return []

        if len(ec) > 0:
            return super()._add(ec, invalidate_cache=invalidate_cache)
        return []

    def satisfiable(self, extra_constraints=(), exact=None):
        try:
            ec = self._constraint_filter(extra_constraints)
            return super().satisfiable(extra_constraints=ec, exact=exact)
        except UnsatError:
            return False

    def eval(self, e, n, extra_constraints=(), exact=None):
        ec = self._constraint_filter(extra_constraints)
        return super().eval(e, n, extra_constraints=ec, exact=exact)

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        ec = self._constraint_filter(extra_constraints)
        return super().batch_eval(exprs, n, extra_constraints=ec, exact=exact)

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        ec = self._constraint_filter(extra_constraints)
        return super().max(e, extra_constraints=ec, signed=signed, exact=exact)

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        ec = self._constraint_filter(extra_constraints)
        return super().min(e, extra_constraints=ec, signed=signed, exact=exact)

    def solution(self, e, v, extra_constraints=(), exact=None):
        ec = self._constraint_filter(extra_constraints)
        return super().solution(e, v, extra_constraints=ec, exact=exact)

    def is_true(self, e, extra_constraints=(), exact=None):
        ec = self._constraint_filter(extra_constraints)
        return super().is_true(e, extra_constraints=ec, exact=exact)

    def is_false(self, e, extra_constraints=(), exact=None):
        ec = self._constraint_filter(extra_constraints)
        return super().is_false(e, extra_constraints=ec, exact=exact)



from __future__ import annotations

from typing import TYPE_CHECKING

from claripy.errors import ClaripyOperationError

if TYPE_CHECKING:
    import claripy


class Annotation:
    """
    Annotations are used to achieve claripy's goal of being an arithmetic instrumentation engine.
    They provide a means to pass extra information to the claripy backends.
    """

    @property
    def eliminatable(self) -> bool:  # pylint:disable=no-self-use
        """
        Returns whether this annotation can be eliminated in a simplification.

        :return: True if eliminatable, False otherwise
        """
        return True

    @property
    def relocatable(self) -> bool:  # pylint:disable=no-self-use
        """
        Returns whether this annotation can be relocated in a simplification.

        :return: True if it can be relocated, false otherwise.
        """
        return False

    def relocate(self, src: claripy.ast.Base, dst: claripy.ast.Base):  # pylint:disable=no-self-use,unused-argument
        """
        This is called when an annotation has to be relocated because of simplifications.

        Consider the following case:

            x = claripy.BVS('x', 32)
            zero = claripy.BVV(0, 32).add_annotation(your_annotation)
            y = x + zero

        Here, one of three things can happen:

            1. if your_annotation.eliminatable is True, the simplifiers will simply
               eliminate your_annotation along with `zero` and `y is x` will hold
            2. elif your_annotation.relocatable is False, the simplifier will abort
               and y will never be simplified
            3. elif your_annotation.relocatable is True, the simplifier will run,
               determine that the simplified result of `x + zero` will be `x`. It
               will then call your_annotation.relocate(zero, x) to move the annotation
               away from the AST that is about to be eliminated.

        :param src: the old AST that was eliminated in the simplification
        :param dst: the new AST (the result of a simplification)
        :return: the annotation that will be applied to `dst`
        """
        if self.relocatable:
            return self
        raise ClaripyOperationError("Annotation is not relocatable")


#
# Some built-in annotations
#


class SimplificationAvoidanceAnnotation(Annotation):
    """SimplificationAvoidanceAnnotation is an annotation that prevents simplification of an AST."""

    @property
    def eliminatable(self):
        return False

    @property
    def relocatable(self):
        return False


class StridedIntervalAnnotation(SimplificationAvoidanceAnnotation):
    """StridedIntervalAnnotation allows annotating a BVS to represent a strided interval."""

    stride: int
    lower_bound: int
    upper_bound: int

    def __init__(self, stride: int, lower_bound: int, upper_bound: int):
        self.stride = stride
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __hash__(self):
        return hash((self.stride, self.lower_bound, self.upper_bound))

    def __eq__(self, other):
        return (
            isinstance(other, StridedIntervalAnnotation)
            and self.stride == other.stride
            and self.lower_bound == other.lower_bound
            and self.upper_bound == other
        )

    def __repr__(self):
        return f"<StridedIntervalAnnotation {self.stride}:{self.lower_bound} - {self.upper_bound}>"


class RegionAnnotation(SimplificationAvoidanceAnnotation):
    """
    Use RegionAnnotation to annotate ASTs. Normally, an AST annotated by
    RegionAnnotations is treated as a ValueSet.
    """

    def __init__(self, region_id, region_base_addr, offset):
        self.region_id = region_id
        self.region_base_addr = region_base_addr
        self.offset = offset

    #
    # Overriding base methods
    #

    def __hash__(self):
        return hash((self.region_id, self.region_base_addr, hash(self.offset)))

    def __repr__(self):
        return f"<RegionAnnotation {self.region_id}:{self.offset:#08x}>"



from __future__ import annotations

from .backend import Backend
from .backend_concrete import BackendConcrete
from .backend_vsa import BackendVSA
from .backend_z3 import BackendZ3

concrete = BackendConcrete()
z3 = BackendZ3()
vsa = BackendVSA()

all_backends = [concrete, z3, vsa]
backends_by_type = {b.__class__.__name__: b for b in all_backends}

__all__ = (
    "Backend",
    "BackendZ3",
    "BackendConcrete",
    "BackendVSA",
    "all_backends",
    "backends_by_type",
    "concrete",
    "z3",
    "vsa",
)



from __future__ import annotations

import logging
import numbers

from claripy.ast.bool import BoolV
from claripy.ast.bv import BV, BVV
from claripy.ast.strings import String, StringV

log = logging.getLogger(__name__)


class Frontend:
    """Frontend is the base class for all claripy Solvers, which are the
    interfaces to the backend constraint solvers.
    """

    def __init__(self):
        pass

    def __getstate__(self):
        return True  # need to return something so that pickle calls setstate

    def __setstate__(self, s):  # pylint:disable=unused-argument
        return

    def branch(self):
        c = self.blank_copy()
        self._copy(c)
        return c

    def blank_copy(self):
        c = self.__class__.__new__(self.__class__)
        self._blank_copy(c)
        return c

    def _blank_copy(self, c):  # pylint:disable=no-self-use,unused-argument
        return

    def _copy(self, c):  # pylint:disable=no-self-use,unused-argument
        return

    #
    # Stuff that should be implemented by subclasses
    #

    def eval_to_ast(self, e, n, extra_constraints=(), exact=None):
        """
        Evaluates expression `e`, returning a list of `n` concrete ASTs.

        :param e:                       the expression
        :param n:                       the number of ASTs to return
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        list of concrete ASTs
        """
        values = self.eval(e, n, extra_constraints=extra_constraints, exact=exact)

        if isinstance(e, BV):
            return [BVV(v, e.size()) for v in values]
        if isinstance(e, String):
            return [StringV(v) for v in values]
        # TODO: Implement support for other types
        raise NotImplementedError

    def finalize(self):
        raise NotImplementedError("finalize() is not implemented")

    def merge(self, others, merge_conditions, common_ancestor=None):
        raise NotImplementedError("merge() is not implemented")

    def combine(self, others):
        raise NotImplementedError("combine() is not implemented")

    def split(self):
        raise NotImplementedError("split() is not implemented")

    def add(self, constraints, invalidate_cache=True):
        """
        Adds constraint(s) to constraints list.

        :param constraints:             constraint(s) to add

        :return:
        """
        constraints = [constraints] if not isinstance(constraints, list | tuple | set) else constraints
        if len(constraints) == 0:
            return []
        constraints = [BoolV(c) if isinstance(c, bool) else c for c in constraints]
        return self._add(constraints, invalidate_cache=invalidate_cache)

    def _add(self, constraints, invalidate_cache=True):
        """
        Adds constraint(s) to constraints list. This version is called by add()
        with constrained constraints.

        :param constraints:             constraint(s) to add

        :return:
        """
        raise NotImplementedError

    def simplify(self):
        """
        Simplifies the stored constraints conjunction.
        """
        raise NotImplementedError

    def check_satisfiability(self, extra_constraints=(), exact=None):
        """
        Checks the satisfiability of stored constraints conjunction.

        :param extra_constraints:       extra constraints to consider when checking satisfiability
        :param exact:                   whether or not to perform exact checking. Ignored by
                                        non-approximating backends.

        :return:                        'SAT' if the conjunction is satisfiable otherwise 'UNSAT'
        """
        raise NotImplementedError

    def satisfiable(self, extra_constraints=(), exact=None):
        """
        Checks if stored constraints conjunction is satisfiable.

        :param extra_constraints:       extra constraints to consider when checking satisfiability
        :param exact:                   whether or not to perform exact checking. Ignored by
                                        non-approximating backends.

        :return:                        True if the conjunction is satisfiable otherwise False
        """
        raise NotImplementedError

    def eval(self, e, n, extra_constraints=(), exact=None):
        """
        Evaluates expression `e`, returning a tuple of `n` solutions.

        :param e:                       the expression
        :param n:                       the number of solutions to return
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        tuple of python primitives representing results
        """
        raise NotImplementedError

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        """
        Evaluates `exprs`, returning a list of tuples (one tuple of `n` solutions for expression).

        :param exprs:                   expressions
        :param n:                       the number of solutions to return
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        list of tuples of python primitives representing results
        """
        raise NotImplementedError

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        """
        Evaluates `e`, returning its max possible value.

        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param signed:                  whether the value should be treated as a signed integer
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        max possible value
        """
        raise NotImplementedError

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        """
        Evaluates `e`, returning its min possible value.

        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param signed:                  whether the value should be treated as a signed integer
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        min possible value
        """
        raise NotImplementedError

    def solution(self, e, v, extra_constraints=(), exact=None):
        """
        Checks if `v` is a possible solution to `e`.

        :param e:                       the expression
        :param v:                       the value
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        True if it is a possible solution otherwise False
        """
        raise NotImplementedError

    def is_true(self, e, extra_constraints=(), exact=None):
        """
        Checks if `e` can only (and TRIVIALLY) evaluate to True. If this function returns True,
        then the expression cannot ever be False, regardless of constraints or anything else.
        If the expression returns False, then the expression might STILL not ever be False; it's just
        that we can't trivially prove it. In other words, a return value of False gives you no
        information whatsoever.


        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        True if it can only evaluate to True otherwise False
        """
        raise NotImplementedError

    def is_false(self, e, extra_constraints=(), exact=None):
        """
        Checks if `e` can only (and TRIVIALLY) evaluate to False. If this function returns True,
        then the expression cannot ever be True, regardless of constraints or anything else.
        If the expression returns False, then the expression might STILL not ever be True; it's just
        that we can't trivially prove it. In other words, a return value of False gives you no
        information whatsoever.

        :param e:                       the expression
        :param extra_constraints:       extra constraints to consider when performing the evaluation
        :param exact:                   whether or not to perform an exact evaluation. Ignored by
                                        non-approximating backends.

        :return:                        True if it can only evaluate to False otherwise False
        """
        raise NotImplementedError

    def downsize(self):  # pylint:disable=no-self-use
        pass

    #
    # Some utility functions
    #

    def _concrete_value(self, e):  # pylint:disable=no-self-use
        if isinstance(e, numbers.Number):
            return e
        return None

    def _concrete_constraint(self, e):  # pylint:disable=no-self-use
        return self._concrete_value(e)



from __future__ import annotations

import re

from claripy.backends.backend_object import BackendObject

from .bv import BVV


class StringV(BackendObject):
    """A concrete string value. Used in the concrete backend for calculations.
    Any use outside of claripy should use `claripy.ast.strings.StringV` instead.
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"StringV({self.value})"


def StrConcat(*args):
    """
    Concatenate a sequence of strings.

    :param args:                    list of string that has to be concatenated

    :return:                        the concatenated string
    """
    new_value = "".join([arg.value for arg in args])
    return StringV(new_value)


def StrSubstr(start_idx, count, initial_string):
    """
    Return the substring of length `count` starting at `start_idx`.

    :param start_idx:               starting index of the substring
    :param count:                   length of the substring in bytes
    :param initial_string:          original string

    :return:                        the substring
    """
    new_value = initial_string.value[start_idx.value : start_idx.value + count.value]
    return StringV(new_value)


def StrReplace(initial_string, pattern_to_be_replaced, replacement_pattern):
    """
    Return string where the first occurrence of `pattern_to_be_replaced` is replaced with
    `replacement_pattern`.

    :param initial_string:          string in which the pattern needs to be replaced
    :param pattern_to_be_replaced:  substring that has to be replaced inside initial_string
    :param replacement_pattern:     pattern that has to be inserted in initial_string to replace
                                    pattern_to_be_replaced

    :return:                        string with replacement
    """
    new_value = initial_string.value.replace(pattern_to_be_replaced.value, replacement_pattern.value, 1)
    return StringV(new_value)


def StrLen(input_string):
    """
    Return length of the `input_string` in bytes.

    :param input_string:            the string we want to calculate the length

    :return:                        bitvector holding the size of the string in bytes
    """
    return BVV(len(input_string.value), 64)


def StrContains(input_string, substring):
    """
    Check if `substring` is contained in `input_string`.

    :param input_string:            the string we want to check
    :param substring:               the string we want to check if it's contained inside the
                                    input_string

    :return:                        True if substring is contained in input_string else False
    """
    return substring.value in input_string.value


def StrPrefixOf(prefix, input_string):
    """
    Check if `input_string` starts with `prefix`.

    :param prefix:                  prefix we want to check
    :param input_string:            the string we want to check

    :return:                        True if the input_string starts with prefix else False
    """
    return re.match(r"^" + prefix.value, input_string.value) is not None


def StrSuffixOf(suffix, input_string):
    """
    Check if `input_string` ends with `suffix`.

    :param suffix:                  suffix we want to check
    :param input_string:            the string we want to check

    :return :                       True if the input_string ends with suffix else False
    """
    return re.match(r".*" + suffix.value + "$", input_string.value) is not None


def StrIndexOf(input_string, substring, startIndex):
    """
    Return the index of the first occurrence of `substring` at or after the `startIndex`, or -1 if
    it is not found.

    :param input_string:            the string we want to check
    :param substring:               the substring we want to find the index
    :param startIndex:              the index to start searching at

    :return BV:                     index of the substring or -1 in bitvector
    """
    try:
        s = input_string.value
        t = substring.value
        i = startIndex.value
        return BVV(i + s[i:].index(t), 64)
    except ValueError:
        return BVV(-1, 64)


def StrToInt(input_string):
    """
    Return the integer representation of `input_string`.

    :param input_string:            the string we want to transform in an integer

    :return BV:                     bitvector of the integer resulting from the string or -1 in
                                    bitvector if the string cannot be transformed into an integer
    """
    try:
        return BVV(int(input_string.value), 64)
    except ValueError:
        return BVV(-1, 64)


def StrIsDigit(input_string):
    """
    Determine whether `input_string` is entirely numeric.

    :param input_string:            the string we want to check

    :return:                        True if the string is entirely numeric otherwise False
    """
    return input_string.value.isdigit()


def IntToStr(input_bvv):
    """
    Return the string representation of `input_bvv`.

    :param input_bvv:               the integer to be expressed as a string

    :return:                        the string representation of the integer
    """
    return StringV(str(input_bvv.value))



from __future__ import annotations

import logging
from itertools import chain
from typing import TypeVar, cast
from weakref import WeakValueDictionary

from claripy.ast.base import Base

log = logging.getLogger(__name__)

T = TypeVar("T", bound=Base)

simplification_cache: WeakValueDictionary[int, Base] = WeakValueDictionary()


def simplify(expr: T) -> T:
    """
    Simplify an expression.
    """

    if expr.is_leaf():
        return expr

    if expr.hash() in simplification_cache and simplification_cache[expr.hash()] is not None:
        return cast(T, simplification_cache[expr.hash()])

    simplified = expr._first_backend("simplify")
    if simplified is None:
        log.debug("Unable to simplify expression")
        simplification_cache[expr.hash()] = expr
        return expr

    # Copy some parameters (that should really go to the Annotation backend)
    simplified._uninitialized = expr.uninitialized

    # dealing with annotations
    if expr.annotations:
        ast_args = tuple(a for a in expr.args if isinstance(a, Base))
        annotations = tuple(
            set(
                chain(
                    chain.from_iterable(a._relocatable_annotations for a in ast_args),
                    tuple(a for a in expr.annotations),
                )
            )
        )
        if annotations != simplified.annotations:
            simplified = simplified.remove_annotations(simplified.annotations)
            simplified = simplified.annotate(*annotations)

    simplification_cache[expr.hash()] = simplified
    return simplified



from __future__ import annotations

from .backend_vsa import BackendVSA
from .bool_result import BoolResult, FalseResult, MaybeResult, TrueResult
from .discrete_strided_interval_set import DEFAULT_MAX_CARDINALITY_WITHOUT_COLLAPSING, DiscreteStridedIntervalSet
from .strided_interval import CreateStridedInterval, StridedInterval
from .valueset import ValueSet

__all__ = (
    "BackendVSA",
    "BoolResult",
    "FalseResult",
    "MaybeResult",
    "TrueResult",
    "DEFAULT_MAX_CARDINALITY_WITHOUT_COLLAPSING",
    "DiscreteStridedIntervalSet",
    "CreateStridedInterval",
    "StridedInterval",
    "ValueSet",
)



from __future__ import annotations

import decimal
from enum import Enum

from claripy.errors import ClaripyOperationError


class RM(Enum):
    """Rounding modes for floating point operations.

    See https://en.wikipedia.org/wiki/IEEE_754#Rounding_rules for more information.
    """

    RM_NearestTiesEven = "RM_RNE"
    RM_NearestTiesAwayFromZero = "RM_RNA"
    RM_TowardsZero = "RM_RTZ"
    RM_TowardsPositiveInf = "RM_RTP"
    RM_TowardsNegativeInf = "RM_RTN"

    @staticmethod
    def default():
        return RM.RM_NearestTiesEven

    def pydecimal_equivalent_rounding_mode(self):
        return {
            RM.RM_TowardsPositiveInf: decimal.ROUND_CEILING,
            RM.RM_TowardsNegativeInf: decimal.ROUND_FLOOR,
            RM.RM_TowardsZero: decimal.ROUND_DOWN,
            RM.RM_NearestTiesEven: decimal.ROUND_HALF_EVEN,
            RM.RM_NearestTiesAwayFromZero: decimal.ROUND_UP,
        }[self]


class FSort:
    """A class representing a floating point sort."""

    def __init__(self, name, exp, mantissa):
        self.name = name
        self.exp = exp
        self.mantissa = mantissa

    def __eq__(self, other):
        return self.exp == other.exp and self.mantissa == other.mantissa

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash((self.name, self.exp, self.mantissa))

    @property
    def length(self):
        return self.exp + self.mantissa

    @staticmethod
    def from_size(n):
        if n == 32:
            return FSORT_FLOAT
        if n == 64:
            return FSORT_DOUBLE
        raise ClaripyOperationError(f"{n} is not a valid FSort size")

    @staticmethod
    def from_params(exp, mantissa):
        if exp == 8 and mantissa == 24:
            return FSORT_FLOAT
        if exp == 11 and mantissa == 53:
            return FSORT_DOUBLE
        raise ClaripyOperationError("unrecognized FSort params")


FSORT_FLOAT = FSort("FLOAT", 8, 24)
FSORT_DOUBLE = FSort("DOUBLE", 11, 53)



from __future__ import annotations

from claripy.frontend import mixin
from claripy.frontend.composite_frontend import CompositeFrontend
from claripy.frontend.frontend import Frontend
from claripy.frontend.full_frontend import FullFrontend
from claripy.frontend.hybrid_frontend import HybridFrontend
from claripy.frontend.light_frontend import LightFrontend
from claripy.frontend.replacement_frontend import ReplacementFrontend

__all__ = (
    "Frontend",
    "CompositeFrontend",
    "FullFrontend",
    "HybridFrontend",
    "LightFrontend",
    "ReplacementFrontend",
    "mixin",
)



from __future__ import annotations

import logging
import numbers
from contextlib import suppress

import claripy
from claripy import backends
from claripy.ast.base import Base
from claripy.ast.bool import BoolV, false
from claripy.ast.bv import BVV
from claripy.errors import BackendError, ClaripyFrontendError

from .constrained_frontend import ConstrainedFrontend

log = logging.getLogger(__name__)


class ReplacementFrontend(ConstrainedFrontend):
    """ReplacementFrontend is a frontend that allows for the replacement
    symbolic constraints with concrete solutions. This is useful for
    simplifying constraints and speeding up solving time.
    """

    # pylint: disable=too-many-positional-arguments

    def __init__(
        self,
        actual_frontend,
        allow_symbolic=None,
        replacements=None,
        replacement_cache=None,
        unsafe_replacement=None,
        complex_auto_replace=None,
        auto_replace=None,
        replace_constraints=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._actual_frontend = actual_frontend
        self._allow_symbolic = True if allow_symbolic is None else allow_symbolic
        self._auto_replace = True if auto_replace is None else auto_replace
        self._complex_auto_replace = False if complex_auto_replace is None else complex_auto_replace
        self._replace_constraints = False if replace_constraints is None else replace_constraints
        self._unsafe_replacement = False if unsafe_replacement is None else unsafe_replacement
        self._replacements = replacements or {}
        self._replacement_cache = replacement_cache or {}

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._actual_frontend = self._actual_frontend.blank_copy()
        c._allow_symbolic = self._allow_symbolic
        c._auto_replace = self._auto_replace
        c._complex_auto_replace = self._complex_auto_replace
        c._replace_constraints = self._replace_constraints
        c._unsafe_replacement = self._unsafe_replacement
        c._replacements = {}
        c._replacement_cache = {}

    def _copy(self, c):
        super()._copy(c)
        self._actual_frontend._copy(c._actual_frontend)

        c._replacements = dict(self._replacements)
        c._replacement_cache = dict(self._replacement_cache)

    #
    # Replacements
    #

    def add_replacement(self, old, new, invalidate_cache=True, replace=True, promote=True):
        if not isinstance(old, Base):
            return

        if old is new:
            return

        if not replace and old.hash() in self._replacements:
            return

        if not promote and old.hash() in self._replacement_cache:
            return

        if not isinstance(new, Base):
            if isinstance(new, bool):
                new = BoolV(new)
            elif isinstance(new, numbers.Number):
                new = BVV(new, old.length)
            else:
                return

        if invalidate_cache:
            self._replacements = dict(self._replacements)
            self._replacement_cache = dict(self._replacements)

        self._replacements[old.hash()] = new
        self._replacement_cache[old.hash()] = new

    def remove_replacements(self, old_entries):
        self._replacements = {k: v for k, v in self._replacements.items() if k not in old_entries}
        self._replacement_cache = dict(self._replacements)

    def clear_replacements(self):
        self._replacements = {}
        self._replacement_cache = dict(self._replacements)

    def _replacement(self, old):
        if not isinstance(old, Base):
            return old

        if old.hash() in self._replacement_cache:
            return self._replacement_cache[old.hash()]

        # not found in the cache
        new = claripy.replace_dict(old, self._replacement_cache)
        if new is not old:
            self._replacement_cache[old.hash()] = new
        return new

    def _add_solve_result(self, e, er, r):
        if not self._auto_replace:
            return
        if not isinstance(e, Base) or not e.symbolic:
            return
        if er.symbolic:
            return
        self.add_replacement(e, r, invalidate_cache=False)

    #
    # Storable support
    #

    def downsize(self):
        self._actual_frontend.downsize()
        self._replacement_cache.clear()

    def __getstate__(self):
        return (
            self._allow_symbolic,
            self._unsafe_replacement,
            self._complex_auto_replace,
            self._auto_replace,
            self._replace_constraints,
            self._replacements,
            self._actual_frontend,
            super().__getstate__(),
        )

    def __setstate__(self, s):
        (
            self._allow_symbolic,
            self._unsafe_replacement,
            self._complex_auto_replace,
            self._auto_replace,
            self._replace_constraints,
            self._replacements,
            self._actual_frontend,
            base_state,
        ) = s

        super().__setstate__(base_state)
        self._replacement_cache = dict(self._replacements)

    #
    # Replacement solving
    #

    def _replace_list(self, lst):
        return tuple(self._replacement(c) for c in lst)

    def eval(self, e, n, extra_constraints=(), exact=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.eval(er, n, extra_constraints=ecr, exact=exact)
        if self._unsafe_replacement:
            self._add_solve_result(e, er, r[0])
        return r

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        er = self._replace_list(exprs)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.batch_eval(er, n, extra_constraints=ecr, exact=exact)
        if self._unsafe_replacement:
            for i, original in enumerate(exprs):
                self._add_solve_result(original, er[i], r[0][i])
        return r

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.max(er, extra_constraints=ecr, signed=signed, exact=exact)
        if self._unsafe_replacement:
            self._add_solve_result(e, er, r)
        return r

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.min(er, extra_constraints=ecr, signed=signed, exact=exact)
        if self._unsafe_replacement:
            self._add_solve_result(e, er, r)
        return r

    def solution(self, e, v, extra_constraints=(), exact=None):
        er = self._replacement(e)
        vr = self._replacement(v)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.solution(er, vr, extra_constraints=ecr, exact=exact)
        if self._unsafe_replacement and r and (not isinstance(vr, Base) or not vr.symbolic):
            self._add_solve_result(e, er, vr)
        return r

    def is_true(self, e, extra_constraints=(), exact=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        return self._actual_frontend.is_true(er, extra_constraints=ecr, exact=exact)

    def is_false(self, e, extra_constraints=(), exact=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        return self._actual_frontend.is_false(er, extra_constraints=ecr, exact=exact)

    def satisfiable(self, extra_constraints=(), exact=None):
        ecr = self._replace_list(extra_constraints)
        return self._actual_frontend.satisfiable(extra_constraints=ecr, exact=exact)

    def _concrete_value(self, e):
        c = super()._concrete_value(e)
        if c is not None:
            return c

        cr = self._replacement(e)
        with suppress(BackendError):
            return backends.concrete.eval(cr, 1)[0]
        return None

    def _concrete_constraint(self, e):
        c = super()._concrete_value(e)
        if c is not None:
            return c

        # if er.is_false():
        #   raise UnsatError("Replacement frontend received False constraint after replacement.")
        if self._replace_constraints:
            er = self._replacement(e)
            return super()._concrete_constraint(er)
        return super()._concrete_constraint(e)

    def _add(self, constraints, invalidate_cache=True):
        if self._auto_replace:
            for c in constraints:
                # the badass thing here would be to use the *replaced* constraint, but
                # we don't currently support chains of replacements, so we'll do a
                # less effective flat-replacement with the original constraint
                # rc = self._replacement(c)
                rc = c
                if not isinstance(rc, Base) or not rc.symbolic:
                    continue

                if not self._complex_auto_replace:
                    if rc.op == "Not":
                        self.add_replacement(c.args[0], false(), replace=False, promote=True, invalidate_cache=True)
                    elif rc.op == "__eq__" and rc.args[0].symbolic ^ rc.args[1].symbolic:
                        old, new = rc.args if rc.args[0].symbolic else rc.args[::-1]
                        self.add_replacement(old, new, replace=False, promote=True, invalidate_cache=True)
                else:
                    satisfiable, replacements = backends.vsa.constraint_to_si(rc)
                    if not satisfiable:
                        self.add_replacement(rc, false())
                    for old, new in replacements:
                        if old.cardinality == 1:
                            continue

                        rold = self._replacement(old)
                        if rold.cardinality == 1:
                            continue

                        self.add_replacement(old, rold.intersection(new))

        added = super()._add(constraints)
        cr = self._replace_list(added)
        if not self._allow_symbolic and any(c.symbolic for c in cr):
            raise ClaripyFrontendError(
                "symbolic constraints made it into ReplacementFrontend with allow_symbolic=False"
            )
        self._actual_frontend.add(cr)

        return added



from __future__ import annotations

import logging
import numbers
import weakref
from contextlib import suppress

from typing_extensions import Self

import claripy
from claripy import operations
from claripy.ast.base import _make_name
from claripy.errors import BackendError, ClaripyValueError
from claripy.util import deprecated

from .bits import Bits
from .bool import Bool, If

log = logging.getLogger(__name__)

_bvv_cache = weakref.WeakValueDictionary()

# pylint: disable=too-many-positional-arguments


class BV(Bits):
    """
    A class representing an AST of operations culminating in a bitvector.
    Do not instantiate this class directly, instead use BVS or BVV to construct a symbol or value, and then use
    operations to construct more complicated expressions.

    Individual sub-bits and bit-ranges can be extracted from a bitvector using index and slice notation.
    Bits are indexed weirdly. For a 32-bit AST:

        a[31] is the *LEFT* most bit, so it'd be the 0 in

            01111111111111111111111111111111

        a[0] is the *RIGHT* most bit, so it'd be the 0 in

            11111111111111111111111111111110

        a[31:30] are the two leftmost bits, so they'd be the 0s in:

            00111111111111111111111111111111

        a[1:0] are the two rightmost bits, so they'd be the 0s in:

            11111111111111111111111111111100
    """

    __slots__ = ()

    def chop(self, bits=1):
        """
        Chops a BV into consecutive sub-slices. Obviously, the length of this BV must be a multiple of bits.

        :returns:   A list of smaller bitvectors, each ``bits`` in length. The first one will be the left-most (i.e.
                    most significant) bits.
        """
        s = len(self)
        if s % bits != 0:
            raise ValueError("expression length (%d) should be a multiple of 'bits' (%d)" % (len(self), bits))
        if s == bits:
            return [self]
        return list(reversed([self[(n + 1) * bits - 1 : n * bits] for n in range(s // bits)]))

    def __getitem__(self, rng):
        if type(rng) is slice:
            left = rng.start if rng.start is not None else len(self) - 1
            right = rng.stop if rng.stop is not None else 0
            if left < 0:
                left = len(self) + left
            if right < 0:
                right = len(self) + right
            return Extract(left, right, self)
        return Extract(int(rng), int(rng), self)

    def get_byte(self, index):
        """
        Extracts a byte from a BV, where the index refers to the byte in a big-endian order

        :param index: the byte to extract
        :return: An 8-bit BV
        """
        return self.get_bytes(index, 1)

    def get_bytes(self, index, size):
        """
        Extracts several bytes from a bitvector, where the index refers to the byte in a big-endian order

        :param index: the byte index at which to start extracting
        :param size: the number of bytes to extract
        :return: A BV of size ``size * 8``
        """
        pos = (self.size() + 7) // 8 - 1 - index
        if pos < 0:
            raise ValueError(
                "Incorrect index %d. Your index must be between %d and %d." % (index, 0, self.size() // 8 - 1)
            )
        if size == 0:
            return BVV(0, 0)
        r = self[min(pos * 8 + 7, self.size() - 1) : (pos - size + 1) * 8]
        if r.size() % 8 != 0:  # pylint:disable=no-member
            r = r.zero_extend(8 - r.size() % 8)  # pylint:disable=no-member
        return r

    def zero_extend(self, n):
        """
        Zero-extends the bitvector by n bits. So:

            a = BVV(0b1111, 4)
            b = a.zero_extend(4)
            b is BVV(0b00001111)
        """
        return ZeroExt(n, self)

    def sign_extend(self, n):
        """
        Sign-extends the bitvector by n bits. So:

            a = BVV(0b1111, 4)
            b = a.sign_extend(4)
            b is BVV(0b11111111)
        """
        return SignExt(n, self)

    def concat(self, *args):
        """
        Concatenates this bitvector with the bitvectors provided.
        This bitvector will be on the far-left, i.e. the most significant bits.
        """
        return Concat(self, *args)

    @staticmethod
    def _from_int(like, value):
        return BVV(value, like.length or 64)

    @staticmethod
    def _from_Bool(like, value):
        return If(value, BVV(1, like.length), BVV(0, like.length))

    @staticmethod
    def _from_bytes(like, value):  # pylint:disable=unused-argument
        return BVV(value)

    @staticmethod
    def _from_str(like, value):  # pylint:disable=unused-argument
        log.warning("BVV value is being coerced from a unicode string, encoding as utf-8")
        value = value.encode("utf-8")
        return BVV(value)

    @staticmethod
    def _from_BVV(like, value):  # pylint:disable=unused-argument
        return BVV(value.value, value.size())

    def val_to_fp(self, sort, signed=True, rm=None):
        """
        Interpret this bitvector as an integer, and return the floating-point representation of that integer.

        :param sort:    The sort of floating point value to return
        :param signed:  Optional: whether this value is a signed integer
        :param rm:      Optional: the rounding mode to use
        :return:        An FP AST whose value is the same as this BV
        """
        if rm is None:
            rm = claripy.fp.RM.default()
        if sort is None:
            sort = claripy.fp.FSort.from_size(self.length)

        op = claripy.ast.fp.fpToFP if signed else claripy.ast.fp.fpToFPUnsigned
        return op(rm, self, sort)

    def raw_to_fp(self):
        """
        Interpret the bits of this bitvector as an IEEE754 floating point number.
        The inverse of this function is raw_to_bv.

        :return:        An FP AST whose bit-pattern is the same as this BV
        """
        sort = claripy.fp.FSort.from_size(self.length)
        return claripy.ast.fp.fpToFP(self, sort)

    def raw_to_bv(self):
        """
        A counterpart to FP.raw_to_bv - does nothing and returns itself.
        """
        return self

    def to_bv(self):
        return self.raw_to_bv()

    def identical(self, other: Self, strict=False) -> bool:
        with suppress(BackendError):
            return claripy.backends.vsa.convert(self).identical(claripy.backends.vsa.convert(other))
        return super().identical(other, strict)


def BVS(  # pylint:disable=redefined-builtin
    name,
    size,
    explicit_name=None,
    **kwargs,
) -> BV:
    """
    Creates a bit-vector symbol (i.e., a variable).

    If you want to specify the maximum or minimum value of a normal symbol that is not part of value-set analysis, you
    should manually add constraints to that effect. **Do not use ``min`` and ``max`` for symbolic execution.**

    :param name:            The name of the symbol.
    :param size:            The size (in bits) of the bit-vector.
    :param bool explicit_name:   If False, an identifier is appended to the name to ensure uniqueness.

    :returns:               a BV object representing this symbol.
    """

    if isinstance(name, bytes):
        name = name.decode()
    if not isinstance(name, str):
        raise TypeError(f"Name value for BVS must be a str, got {type(name)!r}")

    n = _make_name(name, size, False if explicit_name is None else explicit_name)
    encoded_name = n.encode()

    return BV(
        "BVS",
        (n, size),
        variables=frozenset((n,)),
        length=size,
        symbolic=True,
        encoded_name=encoded_name,
        **kwargs,
    )


def BVV(value, size=None, **kwargs) -> BV:
    """
    Creates a bit-vector value (i.e., a concrete value).

    :param value:   The value. Either an integer or a bytestring. If it's the latter, it will be interpreted as the
                    bytes of a big-endian constant.
    :param size:    The size (in bits) of the bit-vector. Optional if you provide a string, required for an integer.

    :returns:       A BV object representing this value.
    """

    if type(value) in (bytes, bytearray, memoryview, str):
        if isinstance(value, str):
            log.warning("BVV value is a unicode string, encoding as utf-8")
            value = value.encode("utf-8")

        if size is None:
            size = len(value) * 8
        elif not isinstance(size, int):
            raise TypeError("Bitvector size  must be either absent (implicit) or an integer")
        elif size != len(value) * 8:
            raise ClaripyValueError("string/size mismatch for BVV creation")

        value = int.from_bytes(value, "big")

    elif size is None or (not isinstance(value, int) and value is not None):
        raise TypeError("BVV() takes either an integer value and a size or a string of bytes")

    # ensure the 0 <= value < (1 << size)
    # FIXME hack to handle None which is used for an Empty Strided Interval (ESI)
    if value is not None:
        value &= (1 << size) - 1

    if not kwargs:
        try:
            return _bvv_cache[(value, size)]
        except KeyError:
            pass

    result = BV("BVV", (value, size), length=size, **kwargs)
    _bvv_cache[(value, size)] = result
    return result


def SI(
    name="unnamed",
    bits=0,
    lower_bound=None,
    upper_bound=None,
    stride=None,
    explicit_name=None,
):
    return BVS(name, bits, explicit_name=explicit_name).annotate(
        claripy.annotation.StridedIntervalAnnotation(stride, lower_bound, upper_bound)
    )


def TSI(bits, name=None, uninitialized=False, explicit_name=None):
    name = "unnamed" if name is None else name
    return BVS(name, bits, uninitialized=uninitialized, explicit_name=explicit_name)


def ESI(bits, **kwargs):
    return BVV(None, bits, **kwargs)


def ValueSet(bits, region=None, region_base_addr=None, value=None, name=None, val=None):
    # Backward compatibility
    if value is None and val is not None:
        value = val
    if region_base_addr is None:
        region_base_addr = 0

    v = region_base_addr + value
    if isinstance(v, claripy.ast.Base):
        v = claripy.simplify(v)

    # Backward compatibility
    if isinstance(v, numbers.Number):
        min_v, max_v = v, v
        stride = 0
    elif isinstance(v, claripy.ast.Base):
        si_anno = v.get_annotation(claripy.annotation.StridedIntervalAnnotation)
        if si_anno is not None:
            min_v = si_anno.lower_bound
            max_v = si_anno.upper_bound
            stride = si_anno.stride
        elif v.op == "BVV":
            min_v = v.args[0]
            max_v = v.args[0]
            stride = 0
        else:
            raise ClaripyValueError(f"ValueSet() does not take `value` ast with op {v.op}")
    else:
        raise ClaripyValueError(f"ValueSet() does not take `value` of type {type(value)}")

    if name is None:
        name = "ValueSet"
    bvs = BVS(name, bits).annotate(
        claripy.annotation.StridedIntervalAnnotation(stride, region_base_addr + min_v, region_base_addr + max_v)
    )

    # Annotate the bvs and return the new AST
    return bvs.annotate(claripy.annotation.RegionAnnotation(region, region_base_addr, value))


VS = ValueSet


#
# Unbound operations
#


# comparisons
ULT = operations.op("__lt__", (BV, BV), Bool, extra_check=operations.length_same_check)
ULE = operations.op("__le__", (BV, BV), Bool, extra_check=operations.length_same_check)
UGT = operations.op("__gt__", (BV, BV), Bool, extra_check=operations.length_same_check)
UGE = operations.op("__ge__", (BV, BV), Bool, extra_check=operations.length_same_check)
SLT = operations.op("SLT", (BV, BV), Bool, extra_check=operations.length_same_check)
SLE = operations.op("SLE", (BV, BV), Bool, extra_check=operations.length_same_check)
SGT = operations.op("SGT", (BV, BV), Bool, extra_check=operations.length_same_check)
SGE = operations.op("SGE", (BV, BV), Bool, extra_check=operations.length_same_check)

# division
SDiv = operations.op(
    "SDiv",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)
SMod = operations.op(
    "SMod",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)

# bit stuff
LShR = operations.op(
    "LShR",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)
SignExt = operations.op(
    "SignExt", (int, BV), BV, extra_check=operations.extend_check, calc_length=operations.ext_length_calc
)
ZeroExt = operations.op(
    "ZeroExt", (int, BV), BV, extra_check=operations.extend_check, calc_length=operations.ext_length_calc
)
Extract = operations.op(
    "Extract",
    (int, int, BV),
    BV,
    extra_check=operations.extract_check,
    calc_length=operations.extract_length_calc,
)

Concat = operations.op("Concat", BV, BV, calc_length=operations.concat_length_calc)

RotateLeft = operations.op(
    "RotateLeft",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)
RotateRight = operations.op(
    "RotateRight",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)
Reverse = operations.op("Reverse", (BV,), BV, calc_length=operations.basic_length_calc)

union = operations.op(
    "union",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)
widen = operations.op(
    "widen",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)
intersection = operations.op(
    "intersection",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)

#
# Bound operations
#

BV.__add__ = operations.op(
    "__add__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__radd__ = operations.reversed_op(BV.__add__)
BV.__floordiv__ = operations.op(
    "__floordiv__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rfloordiv__ = operations.reversed_op(BV.__floordiv__)
BV.__truediv__ = deprecated("BV.__floordiv__", "BV.__truediv__")(BV.__floordiv__)
BV.__rtruediv__ = deprecated("BV.__rfloordiv__", "BV.__rtruediv__")(BV.__rfloordiv__)
BV.__mul__ = operations.op(
    "__mul__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rmul__ = operations.reversed_op(BV.__mul__)
BV.__sub__ = operations.op(
    "__sub__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rsub__ = operations.reversed_op(BV.__sub__)
BV.__pow__ = operations.op(
    "__pow__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rpow__ = operations.reversed_op(BV.__pow__)
BV.__mod__ = operations.op(
    "__mod__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rmod__ = operations.reversed_op(BV.__mod__)
BV.SDiv = operations.op(
    "SDiv",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)
BV.SMod = operations.op(
    "SMod",
    (BV, BV),
    BV,
    extra_check=operations.length_same_check,
    calc_length=operations.basic_length_calc,
)

BV.__neg__ = operations.op("__neg__", (BV,), BV, calc_length=operations.basic_length_calc)
BV.__pos__ = lambda x: x
BV.__abs__ = operations.op("__abs__", (BV,), BV, calc_length=operations.basic_length_calc)

BV.__eq__ = operations.op("__eq__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__ne__ = operations.op("__ne__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__ge__ = operations.op("__ge__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__le__ = operations.op("__le__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__gt__ = operations.op("__gt__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__lt__ = operations.op("__lt__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SLT = operations.op("SLT", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SGT = operations.op("SGT", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SLE = operations.op("SLE", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SGE = operations.op("SGE", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.ULT = operations.op("__lt__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.UGT = operations.op("__gt__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.ULE = operations.op("__le__", (BV, BV), Bool, extra_check=operations.length_same_check)
BV.UGE = operations.op("__ge__", (BV, BV), Bool, extra_check=operations.length_same_check)

BV.__invert__ = operations.op("__invert__", (BV,), BV, calc_length=operations.basic_length_calc)
BV.__or__ = operations.op(
    "__or__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__ror__ = operations.reversed_op(BV.__or__)
BV.__and__ = operations.op(
    "__and__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rand__ = operations.reversed_op(BV.__and__)
BV.__xor__ = operations.op(
    "__xor__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rxor__ = operations.reversed_op(BV.__xor__)
BV.__lshift__ = operations.op(
    "__lshift__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rlshift__ = operations.reversed_op(BV.__lshift__)
BV.__rshift__ = operations.op(
    "__rshift__", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.__rrshift__ = operations.reversed_op(BV.__rshift__)
BV.LShR = operations.op(
    "LShR", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)

BV.Extract = staticmethod(
    operations.op(
        "Extract",
        (int, int, BV),
        BV,
        extra_check=operations.extract_check,
        calc_length=operations.extract_length_calc,
    )
)
BV.Concat = staticmethod(operations.op("Concat", BV, BV, calc_length=operations.concat_length_calc))
BV.reversed = property(operations.op("Reverse", (BV,), BV, calc_length=operations.basic_length_calc))

BV.union = operations.op(
    "union", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.widen = operations.op(
    "widen", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)
BV.intersection = operations.op(
    "intersection", (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc
)



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

from unittest import TestCase

from claripy import BVS, ite_dict


class TestBool(TestCase):
    def test_ite_dict_is_balanced(self):
        case_even = ite_dict(
            BVS("A", 8),
            {
                1: 11,
                2: 22,
                3: 33,
                4: 44,
            },
            BVS("B", 8),
        )

        assert case_even.args[1].depth == case_even.args[2].depth

        case_odd = ite_dict(
            BVS("A", 8),
            {
                1: 11,
                2: 22,
                3: 33,
                4: 44,
                5: 55,
            },
            BVS("B", 8),
        )

        assert case_odd.args[1].depth == case_odd.args[2].depth + 1



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import math
import unittest

import claripy


class TestFp(unittest.TestCase):
    def test_nan(self):
        a = claripy.FPS("a", claripy.FSORT_FLOAT)
        b = claripy.BVS("b", 32)

        s1 = claripy.Solver()
        s1.add((a + 1).isNaN())
        res = s1.eval(a, 1)[0]
        assert math.isnan(res)

        s2 = claripy.Solver()
        s2.add(b.raw_to_fp().isNaN())
        res = s2.eval(b, 1)[0]
        assert res & 0xFF800000 == 0x7F800000 and res & 0x007FFFFF != 0

        s3 = claripy.Solver()
        s3.add(a.isNaN())
        res = s3.eval(a.raw_to_bv(), 1)[0]
        assert res & 0xFF800000 == 0x7F800000 and res & 0x007FFFFF != 0

    def test_nan_set(self):
        v0 = claripy.FPV(float("nan"), claripy.FSORT_DOUBLE)
        v1 = claripy.FPV(float("nan"), claripy.FSORT_DOUBLE)

        assert v0 is v1

        # no exception should be raising
        s = {v0}
        s |= {v1}
        assert len(s) == 1

    def test_negative_zero(self):
        """
        Python does not distinguish between +0.0 and -0.0 and thus, claripy returns same AST for both. However, they
        have different bit representations and hence are different.
        """

        zd = claripy.FPV(0.0, claripy.FSORT_DOUBLE)  # pylint: disable=unused-variable
        nzd = claripy.FPV(-0.0, claripy.FSORT_DOUBLE)
        zf = claripy.FPV(0.0, claripy.FSORT_FLOAT)  # pylint: disable=unused-variable
        nzf = claripy.FPV(-0.0, claripy.FSORT_FLOAT)
        s = claripy.Solver()
        assert s.eval(nzd.to_bv(), 1)[0] == 0x8000000000000000
        assert s.eval(nzf.to_bv(), 1)[0] == 0x80000000

    def test_fp_ops(self):
        a = claripy.FPV(1.5, claripy.FSORT_DOUBLE)
        b = claripy.fpToUBV(claripy.fp.RM.RM_NearestTiesEven, a, 32)

        s = claripy.Solver()
        assert s.eval(b, 1)[0] == 2

    def test_fp_precision_loss(self):
        dd = claripy.FPV(101817237862.0, claripy.FSORT_DOUBLE)
        s = claripy.Solver()
        edd = s.eval(dd.to_bv(), 1)[0]
        edd2 = s.eval(dd.to_fp(claripy.FSORT_FLOAT).to_fp(claripy.FSORT_DOUBLE).to_bv(), 1)[0]
        assert edd != edd2
        assert edd2 == 0x4237B4C7C0000000


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy


class TestAST(unittest.TestCase):
    def test_lite_repr(self):
        one = claripy.BVV(1, 8)
        two = claripy.BVV(2, 8)
        a = claripy.BVS("a", 8, explicit_name=True)
        b = claripy.BVS("b", 8, explicit_name=True)

        assert (a + one * b + two).shallow_repr() == "<BV8 a + 1 * b + 2>"
        assert ((a + one) * (b + two)).shallow_repr() == "<BV8 (a + 1) * (b + 2)>"
        assert (a * one + b * two).shallow_repr() == "<BV8 a * 1 + b * 2>"
        assert (
            (one + a) * (two + b) + (two + a) * (one + b)
        ).shallow_repr() == "<BV8 (1 + a) * (2 + b) + (2 + a) * (1 + b)>"

    def test_associativity(self):
        x = claripy.BVS("x", 8, explicit_name=True)
        y = claripy.BVS("y", 8, explicit_name=True)
        z = claripy.BVS("z", 8, explicit_name=True)
        w = claripy.BVS("w", 8, explicit_name=True)

        assert (x - (y - (z - w))).shallow_repr() == "<BV8 x - (y - (z - w))>"
        assert (x - y - z - w).shallow_repr() == "<BV8 x - y - z - w>"
        assert (x * (y * (z * w))).shallow_repr() == (x * y * z * w).shallow_repr()
        assert (x * y * z * w).shallow_repr() == "<BV8 x * y * z * w>"
        assert (x + y - z - w).shallow_repr() == "<BV8 x + y - z - w>"
        assert (x + (y - (z - w))).shallow_repr() == "<BV8 x + (y - (z - w))>"
        assert (x * y // z % w).shallow_repr() == "<BV8 x * y / z % w>"
        assert (x * (y // (z % w))).shallow_repr() == "<BV8 x * (y / (z % w))>"

    def test_is_leaf(self):
        x = claripy.BVS("x", 8, explicit_name=True)
        y = claripy.BVS("y", 8, explicit_name=True)

        assert x.is_leaf()
        assert y.is_leaf()
        assert (x + y).is_leaf() is False
        assert (x * y).is_leaf() is False
        assert (x - y).is_leaf() is False
        assert (x // y).is_leaf() is False
        assert (x % y).is_leaf() is False
        assert (x**y).is_leaf() is False
        assert (x & y).is_leaf() is False
        assert (x | y).is_leaf() is False
        assert (x ^ y).is_leaf() is False
        assert (x << y).is_leaf() is False
        assert (x >> y).is_leaf() is False
        assert (x == y).is_leaf() is False


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy


class TestReplacements(unittest.TestCase):
    def test_replacement_solver(self):
        sr = claripy.SolverReplacement(claripy.SolverVSA(), replace_constraints=True, complex_auto_replace=True)
        x = claripy.BVS("x", 32)

        sr.add(x + 8 == 10)
        assert sr.max(x) == sr.min(x)

        sr2 = sr.branch()
        sr2.add(x + 8 < 2000)
        assert sr2.max(x) == sr2.min(x) == sr.max(x)

    def test_contradiction(self):
        sr = claripy.SolverReplacement(claripy.Solver(), replace_constraints=True)
        x = claripy.BVS("x", 32)

        sr.add(x == 10)
        assert sr.satisfiable()
        assert sr.eval(x, 10) == (10,)

        sr.add(x == 100)
        assert not sr.satisfiable()

    def test_branching_replacement_solver(self):
        #
        # Simple case: replaceable thing first
        #

        x = claripy.BVS("x", 32)
        s0 = claripy.SolverReplacement(claripy.Solver())
        s0.add(x == 0)

        s1a = s0.branch()
        s1b = s0.branch()

        s1a.add(x == 0)
        s1b.add(x != 0)

        assert s1a.satisfiable()
        assert not s1b.satisfiable()

        #
        # Slightly more complex: different ==
        #

        x = claripy.BVS("x", 32)
        s0 = claripy.SolverReplacement(claripy.Solver())
        s0.add(x == 0)

        s1a = s0.branch()
        s1b = s0.branch()

        s1a.add(x == 0)
        s1b.add(x == 1)

        assert s1a.satisfiable()
        assert not s1b.satisfiable()

        #
        # Complex case: non-replaceable thing first
        #

        # FIXME: Figure this out and uncomment it
        # x = claripy.BVS('x', 32)
        # s0 = claripy.SolverReplacement(claripy.Solver())
        # s0.add(x != 0)
        # s1a = s0.branch()
        # s1b = s0.branch()
        # s1a.add(x != 0)
        # s1b.add(x == 0)
        # assert s1a.satisfiable()
        # assert not s1b.satisfiable()


if __name__ == "__main__":
    unittest.main()
